# -*- coding: utf-8 -*-

"""
The JIT compilation tools for JAX backend.

1. Just-In-Time compilation is implemented by the 'jit()' function

"""

import functools
import logging

import jax

try:
  from jax.errors import UnexpectedTracerError, ConcretizationTypeError
except ImportError:
  from jax.core import UnexpectedTracerError

from brainpy import errors
from brainpy.base.base import Base
from brainpy.base.collector import TensorCollector
from brainpy.math.jaxarray import JaxArray
from brainpy.tools.codes import change_func_name

__all__ = [
  'jit',
]

logger = logging.getLogger('brainpy.math.jit')


def _make_jit(func, vars, static_argnames=None, device=None, f_name=None):
  @functools.partial(jax.jit, static_argnames=static_argnames, device=device)
  def jitted_func(variable_data, *args, **kwargs):
    vars.assign(variable_data)
    out = func(*args, **kwargs)
    changes = vars.dict()
    return out, changes

  def call(*args, **kwargs):
    variable_data = vars.dict()
    try:
      out, changes = jitted_func(variable_data, *args, **kwargs)
    except UnexpectedTracerError as e:
      vars.assign(variable_data)
      raise errors.JaxTracerError(variables=vars) from e
    except ConcretizationTypeError as e:
      vars.assign(variable_data)
      raise errors.ConcretizationTypeError() from e
    vars.assign(changes)
    return out

  return change_func_name(name=f_name, f=call) if f_name else call


def jit(func, dyn_vars=None, static_argnames=None, device=None, auto_infer=True):
  """JIT (Just-In-Time) compilation for class objects.

  This function has the same ability to Just-In-Time compile a pure function,
  but it can also JIT compile a :py:class:`brainpy.DynamicalSystem`, or a
  :py:class:`brainpy.Base` object, or a bounded method for a
  :py:class:`brainpy.Base` object.

  .. note::
    There are several notes when using JIT compilation.

    1. Avoid using scalar in a Variable, TrainVar, etc.

    For example,

    >>> import brainpy as bp
    >>> import brainpy.math as bm
    >>>
    >>> class Test(bp.Base):
    >>>   def __init__(self):
    >>>     super(Test, self).__init__()
    >>>     self.a = bm.Variable(1.)  # Avoid! DO NOT USE!
    >>>   def __call__(self, *args, **kwargs):
    >>>     self.a += 1.

    The above usage is deprecated, because it may cause several errors.
    Instead, we recommend you define the scalar value variable as:

    >>> class Test(bp.Base):
    >>>   def __init__(self):
    >>>     super(Test, self).__init__()
    >>>     self.a = bm.Variable(bm.array([1.]))  # use array to wrap a scalar is recommended
    >>>   def __call__(self, *args, **kwargs):
    >>>     self.a += 1.

    Here, a ndarray is recommended to used to update the variable ``a``.

    2. ``jit`` compilation in ``brainpy.math`` does not support `static_argnums`.
       Instead, users should use `static_argnames`, and call the jitted function with
       keywords like ``jitted_func(arg1=var1, arg2=var2)``. For example,

    >>> def f(a, b, c=1.):
    >>>   if c > 0.: return a + b
    >>>   else: return a * b
    >>>
    >>> # ERROR! https://jax.readthedocs.io/en/latest/notebooks/Common_Gotchas_in_JAX.html#python-control-flow-jit
    >>> bm.jit(f)(1, 2, 0)
    jax._src.errors.ConcretizationTypeError: Abstract tracer value encountered where
    concrete value is expected: Traced<ShapedArray(bool[], weak_type=True)
    >>> # this is right
    >>> bm.jit(f, static_argnames='c')(1, 2, 0)
    DeviceArray(2, dtype=int32, weak_type=True)

  Examples
  --------

  You can JIT a :py:class:`brainpy.DynamicalSystem`

  >>> import brainpy as bp
  >>>
  >>> class LIF(bp.NeuGroup):
  >>>   pass
  >>> lif = bp.math.jit(LIF(10))

  You can JIT a :py:class:`brainpy.Base` object with ``__call__()`` implementation.

  >>> mlp = bp.layers.GRU(100, 200)
  >>> jit_mlp = bp.math.jit(mlp)

  You can also JIT a bounded method of a :py:class:`brainpy.Base` object.

  >>> class Hello(bp.Base):
  >>>   def __init__(self):
  >>>     super(Hello, self).__init__()
  >>>     self.a = bp.math.Variable(bp.math.array(10.))
  >>>     self.b = bp.math.Variable(bp.math.array(2.))
  >>>   def transform(self):
  >>>     return self.a ** self.b
  >>>
  >>> test = Hello()
  >>> bp.math.jit(test.transform)

  Further, you can JIT a normal function, just used like in JAX.

  >>> @bp.math.jit
  >>> def selu(x, alpha=1.67, lmbda=1.05):
  >>>   return lmbda * bp.math.where(x > 0, x, alpha * bp.math.exp(x) - alpha)


  Parameters
  ----------
  func : Base, function, callable
    The instance of Base or a function.
  dyn_vars : optional, dict, tuple, list, JaxArray
    These variables will be changed in the function, or needed in the computation.
  static_argnames : optional, str, list, tuple, dict
    An optional string or collection of strings specifying which named arguments to treat
    as static (compile-time constant). See the comment on ``static_argnums`` for details.
    If not provided but ``static_argnums`` is set, the default is based on calling
    ``inspect.signature(fun)`` to find corresponding named arguments.
  device: optional, Any
    This is an experimental feature and the API is likely to change.
    Optional, the Device the jitted function will run on. (Available devices
    can be retrieved via :py:func:`jax.devices`.) The default is inherited
    from XLA's DeviceAssignment logic and is usually to use
    ``jax.devices()[0]``.
  auto_infer : bool
    Automatical infer the dynamical variables.

  Returns
  -------
  func : Any
    A wrapped version of Base object or function, set up for just-in-time compilation.
  """
  if callable(func):
    if dyn_vars is not None:
      if isinstance(dyn_vars, JaxArray):
        dyn_vars = TensorCollector({'_': dyn_vars})
      elif isinstance(dyn_vars, dict):
        dyn_vars = TensorCollector(dyn_vars)
      elif isinstance(dyn_vars, (tuple, list)):
        dyn_vars = TensorCollector({f'_v{i}': v for i, v in enumerate(dyn_vars)})
      else:
        raise ValueError
    else:
      if auto_infer:
        if isinstance(func, Base):
          dyn_vars = func.vars().unique()
        elif hasattr(func, '__self__') and isinstance(func.__self__, Base):
          dyn_vars = func.__self__.vars().unique()
        else:
          dyn_vars = TensorCollector()
      else:
        dyn_vars = TensorCollector()

    if len(dyn_vars) == 0:  # pure function
      return jax.jit(func,
                     static_argnames=static_argnames,
                     device=device)

    else:  # Base object which implements __call__, or bounded method of Base object
      return _make_jit(vars=dyn_vars,
                       func=func,
                       static_argnames=static_argnames,
                       device=device)

  else:
    raise errors.BrainPyError(f'Only support instance of {Base.__name__}, or a callable '
                              f'function, but we got {type(func)}.')



# -*- coding: utf-8 -*-

import brainpy.math as bm
from brainpy.base import Base

__all__ = [
  # base class for offline training algorithm
  'OfflineAlgorithm',

  # training methods
  'RidgeRegression',
  'LinearRegression',

  # general supports
  'get_supported_offline_methods',
  'register_offline_method',
]

name2func = dict()


class OfflineAlgorithm(Base):
  """Base class for offline training algorithm."""

  def __init__(self, name=None):
    super(OfflineAlgorithm, self).__init__(name=name)

  def __call__(self, targets, inputs, outputs):
    """The training procedure.

    Parameters
    ----------
    inputs: JaxArray, jax.numpy.ndarray, numpy.ndarray
      The 3d input data with the shape of `(num_batch, num_time, num_input)`,
      or, the 2d input data with the shape of `(num_time, num_input)`.

    targets: JaxArray, jax.numpy.ndarray, numpy.ndarray
      The 3d target data with the shape of `(num_batch, num_time, num_output)`,
      or the 2d target data with the shape of `(num_time, num_output)`.

    outputs: JaxArray, jax.numpy.ndarray, numpy.ndarray
      The 3d output data with the shape of `(num_batch, num_time, num_output)`,
      or the 2d output data with the shape of `(num_time, num_output)`.

    Returns
    -------
    weight: JaxArray
      The weights after fit.
    """
    raise NotImplementedError('Must implement the __call__ function by the subclass itself.')

  def __repr__(self):
    return self.__class__.__name__


class RidgeRegression(OfflineAlgorithm):
  """Training algorithm of ridge regression.

  Parameters
  ----------
  beta: float
    The regularization coefficient.
  """

  def __init__(self, beta=1e-7, name=None):
    super(RidgeRegression, self).__init__(name=name)
    self.beta = beta

  def __call__(self, targets, inputs, outputs=None):
    # checking
    inputs = bm.asarray(inputs).reshape((-1, inputs.shape[2]))
    targets = bm.asarray(targets).reshape((-1, targets.shape[2]))
    # solving
    temp = inputs.T @ inputs
    if self.beta > 0.:
      temp += self.beta * bm.eye(inputs.shape[-1])
    weights = bm.linalg.pinv(temp) @ (inputs.T @ targets)
    return weights

  def __repr__(self):
    return f'{self.__class__.__name__}(beta={self.beta})'


name2func['ridge'] = RidgeRegression


class LinearRegression(OfflineAlgorithm):
  """Training algorithm of least-square regression."""

  def __init__(self, name=None):
    super(LinearRegression, self).__init__(name=name)

  def __call__(self, targets, inputs, outputs=None):
    inputs = bm.asarray(inputs).reshape((-1, inputs.shape[2]))
    targets = bm.asarray(targets).reshape((-1, targets.shape[2]))
    weights = bm.linalg.lstsq(inputs, targets)
    return weights[0]


name2func['linear'] = LinearRegression
name2func['lstsq'] = LinearRegression


class LassoRegression(OfflineAlgorithm):
  """Lasso regression method for offline training.

  Parameters
  ----------
  alpha: float
    Constant that multiplies the L1 term. Defaults to 1.0.
    `alpha = 0` is equivalent to an ordinary least square.
  max_iter: int
    The maximum number of iterations.
  """

  def __init__(self, alpha=1.0, max_iter=1000, name=None):
    super(LassoRegression, self).__init__(name=name)
    self.alpha = alpha
    self.max_iter = max_iter

  def __call__(self, *args, **kwargs):
    pass


# name2func['lasso'] = LassoRegression


def elastic_net_regression(x, y, train_pars):
  pass


# name2func['elastic_net'] = elastic_net_regression


def logistic_regression(x, y, train_pars):
  pass


# name2func['logistic'] = logistic_regression


def polynomial_regression(x, y, train_pars):
  pass


# name2func['polynomial'] = polynomial_regression


def stepwise_regression(x, y, train_pars):
  pass


# name2func['stepwise'] = stepwise_regression


def get_supported_offline_methods():
  """Get all supported offline training methods."""
  return tuple(name2func.keys())


def register_offline_method(name, method):
  """Register a new offline learning method.

  Parameters
  ----------
  name: str
    The method name.
  method: callable
    The function method.
  """
  if name in name2func:
    raise ValueError(f'"{name}" has been registered in offline training methods.')
  if not callable(method):
    raise ValueError(f'"method" must be an instance of callable '
                     f'function, but we got {type(method)}')
  name2func[name] = method


def get(name):
  """Get the training function according to the training method name."""
  if name not in name2func:
    raise ValueError(f'All offline methods are: {get_supported_offline_methods()}.\n'
                     f'But we got {name}.')
  return name2func[name]



# -*- coding: utf-8 -*-

from typing import TypeVar, Tuple

import jax.numpy as jnp

import brainpy.math as bm

__all__ = [
  'Tensor',
  'Parameter',
  'Shape',
]

Tensor = TypeVar('Tensor', bm.JaxArray, jnp.ndarray)
Parameter = TypeVar('Parameter', float, int, jnp.ndarray, bm.JaxArray, bm.Variable)
Shape = TypeVar('Shape', int, Tuple[int, ...])



# -*- coding: utf-8 -*-

from itertools import combinations_with_replacement
from typing import Union, Sequence

import numpy as np
import jax.numpy as jnp

import brainpy.math as bm
from brainpy.nn.base import RecurrentNode
from brainpy.nn.datatypes import MultipleData
from brainpy.tools.checking import (check_shape_consistency,
                                    check_integer,
                                    check_sequence)

__all__ = [
  'NVAR'
]


def _comb(N, k):
  r"""The number of combinations of N things taken k at a time.

  .. math::

     \frac{N!}{(N-k)! k!}

  """
  if N > k:
    val = 1
    for j in range(min(k, N - k)):
      val = (val * (N - j)) // (j + 1)
    return val
  elif N == k:
    return 1
  else:
    return 0


class NVAR(RecurrentNode):
  """Nonlinear vector auto-regression (NVAR) node.

  This class has the following features:

  - it supports batch size,
  - it supports multiple orders,

  Parameters
  ----------
  delay: int
    The number of delay step.
  order: int, sequence of int
    The nonlinear order.
  stride: int
    The stride to sample linear part vector in the delays.
  constant: optional, float
    The constant value.

  References
  ----------
  .. [1] Gauthier, D.J., Bollt, E., Griffith, A. et al. Next generation
         reservoir computing. Nat Commun 12, 5564 (2021).
         https://doi.org/10.1038/s41467-021-25801-2

  """
  data_pass = MultipleData('sequence')

  def __init__(
      self,
      delay: int,
      order: Union[int, Sequence[int]] = None,
      stride: int = 1,
      constant: bool = False,
      trainable: bool = False,
      **kwargs
  ):
    super(NVAR, self).__init__(trainable=trainable, **kwargs)

    # parameters
    order = tuple() if order is None else order
    if not isinstance(order, (tuple, list)):
      order = (order,)
    self.order = tuple(order)
    check_sequence(order, 'order', allow_none=False)
    for o in order: check_integer(o, 'delay', allow_none=False, min_bound=2)
    check_integer(delay, 'delay', allow_none=False, min_bound=1)
    check_integer(stride, 'stride', allow_none=False, min_bound=1)
    assert isinstance(constant, bool), f'Must be an instance of boolean, but got {constant}.'
    self.delay = delay
    self.stride = stride
    self.constant = constant
    self.num_delay = 1 + (self.delay - 1) * self.stride

    # attributes
    self.comb_ids = []
    self.feature_names = []
    self.input_dim = None
    self.output_dim = None
    self.linear_dim = None
    self.nonlinear_dim = None

    # delay variables
    self.idx = bm.Variable(jnp.asarray([0], dtype=bm.int_))
    self.store = None

  def init_ff_conn(self):
    """Initialize feedforward connections."""
    # input dimension
    batch_size, free_size = check_shape_consistency(self.feedforward_shapes, -1, True)
    self.input_dim = sum(free_size)
    assert batch_size == (None,), f'batch_size must be None, but got {batch_size}'
    # linear dimension
    self.linear_dim = self.delay * self.input_dim
    # For each monomial created in the non-linear part, indices
    # of the n components involved, n being the order of the
    # monomials. Precompute them to improve efficiency.
    for order in self.order:
      assert order >= 2, f'"order" must be a integer >= 2, while we got {order}.'
      idx = np.array(list(combinations_with_replacement(np.arange(self.linear_dim), order)))
      self.comb_ids.append(jnp.asarray(idx))
    # number of non-linear components is (d + n - 1)! / (d - 1)! n!
    # i.e. number of all unique monomials of order n made from the
    # linear components.
    self.nonlinear_dim = sum([len(ids) for ids in self.comb_ids])
    # output dimension
    self.output_dim = int(self.linear_dim + self.nonlinear_dim)
    if self.constant:
      self.output_dim += 1
    self.set_output_shape((None, self.output_dim))

  def init_state(self, num_batch=1):
    """Initialize the node state which depends on batch size."""
    # To store the last inputs.
    # Note, the batch axis is not in the first dimension, so we
    # manually handle the state of NVAR, rather return it.
    state = jnp.zeros((self.num_delay, num_batch, self.input_dim), dtype=bm.float_)
    if self.store is None:
      self.store = bm.Variable(state)
    else:
      self.store._value = state

  def forward(self, ff, fb=None, **shared_kwargs):
    all_parts = []
    # 1. Store the current input
    ff = bm.concatenate(ff, axis=-1)
    self.store[self.idx[0]] = ff
    # 2. Linear part:
    # select all previous inputs, including the current, with strides
    select_ids = (self.idx[0] - jnp.arange(0, self.num_delay, self.stride)) % self.num_delay
    linear_parts = jnp.moveaxis(self.store[select_ids], 0, 1)  # (num_batch, num_time, num_feature)
    linear_parts = jnp.reshape(linear_parts, (linear_parts.shape[0], -1))
    # 3. constant
    if self.constant:
      constant = jnp.ones((linear_parts.shape[0], 1), dtype=ff.dtype)
      all_parts.append(constant)
    all_parts.append(linear_parts)
    # 3. Nonlinear part:
    # select monomial terms and compute them
    for ids in self.comb_ids:
      all_parts.append(jnp.prod(linear_parts[:, ids], axis=2))
    # 4. Finally
    self.idx.value = (self.idx + 1) % self.num_delay
    return jnp.concatenate(all_parts, axis=-1)

  def get_feature_names(self):
    """Get output feature names for transformation.

    Returns
    -------
    feature_names_out : list of str
        Transformed feature names.
    """
    if not self.is_initialized:
      raise ValueError('Please initialize the node first.')
    linear_names = [f'x{i}(t)' for i in range(self.input_dim)]
    for di in range(1, self.delay):
      linear_names.extend([f'x{i}(t-{di * self.stride})' for i in range(self.input_dim)])
    nonlinear_names = []
    for ids in self.comb_ids:
      for id_ in np.asarray(ids):
        uniques, counts = np.unique(id_, return_counts=True)
        nonlinear_names.append(" ".join(
          "%s^%d" % (linear_names[ind], exp) if (exp != 1) else linear_names[ind]
          for ind, exp in zip(uniques, counts)
        ))
    all_names = linear_names + nonlinear_names
    if self.constant:
      all_names = ['1'] + all_names
    return all_names

  def get_feature_names_for_plot(self):
    """Get output feature names for matplotlib plotting.

    Returns
    -------
    feature_names_out : list of str
        Transformed feature names.
    """
    if not self.is_initialized:
      raise ValueError('Please initialize the node first.')
    linear_names = [f'x{i}_t' for i in range(self.input_dim)]
    for di in range(1, self.delay):
      linear_names.extend([(f'x{i}_' + r'{t-%d}' % (di * self.stride))
                           for i in range(self.input_dim)])
    nonlinear_names = []
    for ids in self.comb_ids:
      for id_ in np.asarray(ids):
        uniques, counts = np.unique(id_, return_counts=True)
        nonlinear_names.append(" ".join(
          "%s^%d" % (linear_names[ind], exp) if (exp != 1) else linear_names[ind]
          for ind, exp in zip(uniques, counts)
        ))
    all_names = [f'${n}$' for n in linear_names] + [f'${n}$' for n in nonlinear_names]
    if self.constant:
      all_names = ['1'] + all_names
    return all_names



# -*- coding: utf-8 -*-


from typing import Union, Callable

import brainpy.math as bm
from brainpy.initialize import (XavierNormal,
                                ZeroInit,
                                Uniform,
                                Orthogonal,
                                init_param,
                                Initializer)
from brainpy.nn.base import RecurrentNode
from brainpy.nn.datatypes import MultipleData
from brainpy.tools.checking import (check_integer,
                                    check_initializer,
                                    check_shape_consistency)
from brainpy.types import Tensor

__all__ = [
  'VanillaRNN',
  'GRU',
  'LSTM',
]


class VanillaRNN(RecurrentNode):
  r"""Basic fully-connected RNN core.

  Given :math:`x_t` and the previous hidden state :math:`h_{t-1}` the
  core computes

  .. math::

     h_t = \mathrm{ReLU}(w_i x_t + b_i + w_h h_{t-1} + b_h)

  The output is equal to the new state, :math:`h_t`.


  Parameters
  ----------
  num_unit: int
    The number of hidden unit in the node.
  state_initializer: callable, Initializer, bm.ndarray, jax.numpy.ndarray
    The state initializer.
  wi_initializer: callable, Initializer, bm.ndarray, jax.numpy.ndarray
    The input weight initializer.
  wh_initializer: callable, Initializer, bm.ndarray, jax.numpy.ndarray
    The hidden weight initializer.
  bias_initializer: optional, callable, Initializer, bm.ndarray, jax.numpy.ndarray
    The bias weight initializer.
  activation: str, callable
    The activation function. It can be a string or a callable function.
    See ``brainpy.math.activations`` for more details.
  trainable: bool
    Whether set the node is trainable.

  """
  data_pass = MultipleData('sequence')

  def __init__(
      self,
      num_unit: int,
      state_initializer: Union[Tensor, Callable, Initializer] = Uniform(),
      wi_initializer: Union[Tensor, Callable, Initializer] = XavierNormal(),
      wh_initializer: Union[Tensor, Callable, Initializer] = XavierNormal(),
      bias_initializer: Union[Tensor, Callable, Initializer] = ZeroInit(),
      activation: str = 'relu',
      **kwargs
  ):
    super(VanillaRNN, self).__init__(**kwargs)

    self.num_unit = num_unit
    check_integer(num_unit, 'num_unit', min_bound=1, allow_none=False)
    self.set_output_shape((None, self.num_unit))

    # initializers
    self._state_initializer = state_initializer
    self._wi_initializer = wi_initializer
    self._wh_initializer = wh_initializer
    self._bias_initializer = bias_initializer
    check_initializer(wi_initializer, 'wi_initializer', allow_none=False)
    check_initializer(wh_initializer, 'wh_initializer', allow_none=False)
    check_initializer(state_initializer, 'state_initializer', allow_none=False)
    check_initializer(bias_initializer, 'bias_initializer', allow_none=True)

    # activation function
    self.activation = bm.activations.get(activation)

  def init_ff_conn(self):
    unique_size, free_sizes = check_shape_consistency(self.feedforward_shapes, -1, True)
    assert len(unique_size) == 1, 'Only support data with or without batch size.'
    # weights
    num_input = sum(free_sizes)
    self.Wff = init_param(self._wi_initializer, (num_input, self.num_unit))
    self.Wrec = init_param(self._wh_initializer, (self.num_unit, self.num_unit))
    self.bias = init_param(self._bias_initializer, (self.num_unit,))
    if self.trainable:
      self.Wff = bm.TrainVar(self.Wff)
      self.Wrec = bm.TrainVar(self.Wrec)
      self.bias = None if (self.bias is None) else bm.TrainVar(self.bias)

  def init_fb_conn(self):
    unique_size, free_sizes = check_shape_consistency(self.feedback_shapes, -1, True)
    assert len(unique_size) == 1, 'Only support data with or without batch size.'
    num_feedback = sum(free_sizes)
    # weights
    self.Wfb = init_param(self._wi_initializer, (num_feedback, self.num_unit))
    if self.trainable:
      self.Wfb = bm.TrainVar(self.Wfb)

  def init_state(self, num_batch=1):
    return init_param(self._state_initializer, (num_batch, self.num_unit))

  def forward(self, ff, fb=None, **shared_kwargs):
    ff = bm.concatenate(ff, axis=-1)
    h = ff @ self.Wff
    h += self.state.value @ self.Wrec
    if self.bias is not None:
      h += self.bias
    if fb is not None:
      fb = bm.concatenate(fb, axis=-1)
      h += fb @ self.Wfb
    self.state.value = self.activation(h)
    return self.state.value


class GRU(RecurrentNode):
  r"""Gated Recurrent Unit.

  The implementation is based on (Chung, et al., 2014) [1]_ with biases.

  Given :math:`x_t` and the previous state :math:`h_{t-1}` the core computes

  .. math::

     \begin{array}{ll}
     z_t &= \sigma(W_{iz} x_t + W_{hz} h_{t-1} + b_z) \\
     r_t &= \sigma(W_{ir} x_t + W_{hr} h_{t-1} + b_r) \\
     a_t &= \tanh(W_{ia} x_t + W_{ha} (r_t \bigodot h_{t-1}) + b_a) \\
     h_t &= (1 - z_t) \bigodot h_{t-1} + z_t \bigodot a_t
     \end{array}

  where :math:`z_t` and :math:`r_t` are reset and update gates.

  The output is equal to the new hidden state, :math:`h_t`.

  Warning: Backwards compatibility of GRU weights is currently unsupported.

  Parameters
  ----------
  num_unit: int
    The number of hidden unit in the node.
  state_initializer: callable, Initializer, bm.ndarray, jax.numpy.ndarray
    The state initializer.
  wi_initializer: callable, Initializer, bm.ndarray, jax.numpy.ndarray
    The input weight initializer.
  wh_initializer: callable, Initializer, bm.ndarray, jax.numpy.ndarray
    The hidden weight initializer.
  bias_initializer: optional, callable, Initializer, bm.ndarray, jax.numpy.ndarray
    The bias weight initializer.
  activation: str, callable
    The activation function. It can be a string or a callable function.
    See ``brainpy.math.activations`` for more details.
  trainable: bool
    Whether set the node is trainable.

  References
  ----------
  .. [1] Chung, J., Gulcehre, C., Cho, K. and Bengio, Y., 2014. Empirical
         evaluation of gated recurrent neural networks on sequence modeling.
         arXiv preprint arXiv:1412.3555.
  """
  data_pass = MultipleData('sequence')

  def __init__(
      self,
      num_unit: int,
      wi_initializer: Union[Tensor, Callable, Initializer] = Orthogonal(),
      wh_initializer: Union[Tensor, Callable, Initializer] = Orthogonal(),
      bias_initializer: Union[Tensor, Callable, Initializer] = ZeroInit(),
      state_initializer: Union[Tensor, Callable, Initializer] = ZeroInit(),
      **kwargs
  ):
    super(GRU, self).__init__(**kwargs)

    self.num_unit = num_unit
    check_integer(num_unit, 'num_unit', min_bound=1, allow_none=False)
    self.set_output_shape((None, self.num_unit))

    self._wi_initializer = wi_initializer
    self._wh_initializer = wh_initializer
    self._bias_initializer = bias_initializer
    self._state_initializer = state_initializer
    check_initializer(wi_initializer, 'wi_initializer', allow_none=False)
    check_initializer(wh_initializer, 'wh_initializer', allow_none=False)
    check_initializer(state_initializer, 'state_initializer', allow_none=False)
    check_initializer(bias_initializer, 'bias_initializer', allow_none=True)

  def init_ff_conn(self):
    # data shape
    unique_size, free_sizes = check_shape_consistency(self.feedforward_shapes, -1, True)
    assert len(unique_size) == 1, 'Only support data with or without batch size.'

    # weights
    num_input = sum(free_sizes)
    self.Wi_ff = init_param(self._wi_initializer, (num_input, self.num_unit * 3))
    self.Wh = init_param(self._wh_initializer, (self.num_unit, self.num_unit * 3))
    self.bias = init_param(self._bias_initializer, (self.num_unit * 3,))
    if self.trainable:
      self.Wi_ff = bm.TrainVar(self.Wi_ff)
      self.Wh = bm.TrainVar(self.Wh)
      self.bias = bm.TrainVar(self.bias) if (self.bias is not None) else None

  def init_fb_conn(self):
    unique_size, free_sizes = check_shape_consistency(self.feedback_shapes, -1, True)
    assert len(unique_size) == 1, 'Only support data with or without batch size.'
    num_feedback = sum(free_sizes)
    # weights
    self.Wi_fb = init_param(self._wi_initializer, (num_feedback, self.num_unit * 3))
    if self.trainable:
      self.Wi_fb = bm.TrainVar(self.Wi_fb)

  def init_state(self, num_batch=1):
    return init_param(self._state_initializer, (num_batch, self.num_unit))

  def forward(self, ff, fb=None, **shared_kwargs):
    gates_x = bm.matmul(bm.concatenate(ff, axis=-1), self.Wi_ff)
    if fb is not None:
      gates_x += bm.matmul(bm.concatenate(fb, axis=-1), self.Wi_fb)
    zr_x, a_x = bm.split(gates_x, indices_or_sections=[2 * self.num_unit], axis=-1)
    w_h_z, w_h_a = bm.split(self.Wh, indices_or_sections=[2 * self.num_unit], axis=-1)
    zr_h = bm.matmul(self.state, w_h_z)
    zr = zr_x + zr_h
    has_bias = (self.bias is not None)
    if has_bias:
      b_z, b_a = bm.split(self.bias, indices_or_sections=[2 * self.num_unit], axis=0)
      zr += bm.broadcast_to(b_z, zr_h.shape)
    z, r = bm.split(bm.sigmoid(zr), indices_or_sections=2, axis=-1)
    a_h = bm.matmul(r * self.state, w_h_a)
    if has_bias:
      a = bm.tanh(a_x + a_h + bm.broadcast_to(b_a, a_h.shape))
    else:
      a = bm.tanh(a_x + a_h)
    next_state = (1 - z) * self.state + z * a
    self.state.value = next_state
    return next_state


class LSTM(RecurrentNode):
  r"""Long short-term memory (LSTM) RNN core.

  The implementation is based on (zaremba, et al., 2014) [1]_. Given
  :math:`x_t` and the previous state :math:`(h_{t-1}, c_{t-1})` the core
  computes

  .. math::

     \begin{array}{ll}
     i_t = \sigma(W_{ii} x_t + W_{hi} h_{t-1} + b_i) \\
     f_t = \sigma(W_{if} x_t + W_{hf} h_{t-1} + b_f) \\
     g_t = \tanh(W_{ig} x_t + W_{hg} h_{t-1} + b_g) \\
     o_t = \sigma(W_{io} x_t + W_{ho} h_{t-1} + b_o) \\
     c_t = f_t c_{t-1} + i_t g_t \\
     h_t = o_t \tanh(c_t)
     \end{array}

  where :math:`i_t`, :math:`f_t`, :math:`o_t` are input, forget and
  output gate activations, and :math:`g_t` is a vector of cell updates.

  The output is equal to the new hidden, :math:`h_t`.

  Notes
  -----

  Forget gate initialization: Following (Jozefowicz, et al., 2015) [2]_ we add 1.0
  to :math:`b_f` after initialization in order to reduce the scale of forgetting in
  the beginning of the training.


  Parameters
  ----------
  num_unit: int
    The number of hidden unit in the node.
  state_initializer: callable, Initializer, bm.ndarray, jax.numpy.ndarray
    The state initializer.
  wi_initializer: callable, Initializer, bm.ndarray, jax.numpy.ndarray
    The input weight initializer.
  wh_initializer: callable, Initializer, bm.ndarray, jax.numpy.ndarray
    The hidden weight initializer.
  bias_initializer: optional, callable, Initializer, bm.ndarray, jax.numpy.ndarray
    The bias weight initializer.
  activation: str, callable
    The activation function. It can be a string or a callable function.
    See ``brainpy.math.activations`` for more details.
  trainable: bool
    Whether set the node is trainable.

  References
  ----------

  .. [1] Zaremba, Wojciech, Ilya Sutskever, and Oriol Vinyals. "Recurrent neural
         network regularization." arXiv preprint arXiv:1409.2329 (2014).
  .. [2] Jozefowicz, Rafal, Wojciech Zaremba, and Ilya Sutskever. "An empirical
         exploration of recurrent network architectures." In International conference
         on machine learning, pp. 2342-2350. PMLR, 2015.
  """
  data_pass = MultipleData('sequence')

  def __init__(
      self,
      num_unit: int,
      wi_initializer: Union[Tensor, Callable, Initializer] = XavierNormal(),
      wh_initializer: Union[Tensor, Callable, Initializer] = XavierNormal(),
      bias_initializer: Union[Tensor, Callable, Initializer] = ZeroInit(),
      state_initializer: Union[Tensor, Callable, Initializer] = ZeroInit(),
      **kwargs
  ):
    super(LSTM, self).__init__(**kwargs)

    self.num_unit = num_unit
    check_integer(num_unit, 'num_unit', min_bound=1, allow_none=False)
    self.set_output_shape((None, self.num_unit,))

    self._state_initializer = state_initializer
    self._wi_initializer = wi_initializer
    self._wh_initializer = wh_initializer
    self._bias_initializer = bias_initializer
    check_initializer(wi_initializer, 'wi_initializer', allow_none=False)
    check_initializer(wh_initializer, 'wh_initializer', allow_none=False)
    check_initializer(bias_initializer, 'bias_initializer', allow_none=True)
    check_initializer(state_initializer, 'state_initializer', allow_none=False)

  def init_ff_conn(self):
    # data shape
    unique_size, free_sizes = check_shape_consistency(self.feedforward_shapes, -1, True)
    assert len(unique_size) == 1, 'Only support data with or without batch size.'
    # weights
    num_input = sum(free_sizes)
    self.Wi_ff = init_param(self._wi_initializer, (num_input, self.num_unit * 4))
    self.Wh = init_param(self._wh_initializer, (self.num_unit, self.num_unit * 4))
    self.bias = init_param(self._bias_initializer, (self.num_unit * 4,))
    if self.trainable:
      self.Wi_ff = bm.TrainVar(self.Wi_ff)
      self.Wh = bm.TrainVar(self.Wh)
      self.bias = None if (self.bias is None) else bm.TrainVar(self.bias)

  def init_fb_conn(self):
    unique_size, free_sizes = check_shape_consistency(self.feedback_shapes, -1, True)
    assert len(unique_size) == 1, 'Only support data with or without batch size.'
    num_feedback = sum(free_sizes)
    # weights
    self.Wi_fb = init_param(self._wi_initializer, (num_feedback, self.num_unit * 4))
    if self.trainable:
      self.Wi_fb = bm.TrainVar(self.Wi_fb)

  def init_state(self, num_batch=1):
    return init_param(self._state_initializer, (num_batch * 2, self.num_unit))

  def forward(self, ff, fb=None, **shared_kwargs):
    h, c = bm.split(self.state, 2)
    gated = bm.concatenate(ff, axis=-1) @ self.Wi_ff
    if fb is not None:
      gated += bm.concatenate(fb, axis=-1) @ self.Wi_fb
    if self.bias is not None:
      gated += self.bias
    gated += h @ self.Wh
    i, g, f, o = bm.split(gated, indices_or_sections=4, axis=-1)
    c = bm.sigmoid(f + 1.) * c + bm.sigmoid(i) * bm.tanh(g)
    h = bm.sigmoid(o) * bm.tanh(c)
    self.state.value = bm.vstack([h, c])
    return h

  @property
  def h(self):
    """Hidden state."""
    return bm.split(self.state, 2)[0]

  @h.setter
  def h(self, value):
    if self.state is None:
      raise ValueError('Cannot set "h" state. Because the state is not initialized.')
    self.state[:self.state.shape[0] // 2, :] = value

  @property
  def c(self):
    """Memory cell."""
    return bm.split(self.state, 2)[1]

  @c.setter
  def c(self, value):
    if self.state is None:
      raise ValueError('Cannot set "c" state. Because the state is not initialized.')
    self.state[self.state.shape[0] // 2:, :] = value


class ConvNDLSTM(RecurrentNode):
  data_pass = MultipleData('sequence')


class Conv1DLSTM(ConvNDLSTM):
  data_pass = MultipleData('sequence')


class Conv2DLSTM(ConvNDLSTM):
  data_pass = MultipleData('sequence')


class Conv3DLSTM(ConvNDLSTM):
  data_pass = MultipleData('sequence')



# -*- coding: utf-8 -*-

import time
from typing import Union, Dict, Callable, Sequence

import jax.numpy as jnp
import numpy as np
from jax import jit, random as jr
from jax.tree_util import tree_map

import brainpy.losses as losses
import brainpy.math as bm
import brainpy.optimizers as optim
from brainpy.errors import UnsupportedError
from brainpy.nn.base import Node, Network
from brainpy.nn.utils import check_data_batch_size, serialize_kwargs
from brainpy.tools.checking import check_dict_data, check_float
from brainpy.types import Tensor
from .rnn_trainer import RNNTrainer

__all__ = [
  'BPTT',
  'BPFF',
]


class BPTT(RNNTrainer):
  """
  The trainer implementing back propagation through time (BPTT)
  algorithm for recurrent neural networks.

  """

  def __init__(
      self,
      target: Node,

      # arguments for BPTT trainer
      loss: Union[str, Callable],  # loss function
      optimizer: optim.Optimizer = None,  # optimizer
      max_grad_norm=None,
      shuffle_data: bool = True,
      jit: bool = True,

      # common arguments for RNNTrainer
      **kwargs
  ):
    super(BPTT, self).__init__(target=target, **kwargs)

    # jit settings
    if isinstance(jit, bool):
      self.jit = {'fit': jit, 'predict': jit, 'loss': jit}
    elif isinstance(jit, dict):
      jit = {key: val for key, val in jit.items()}
      self.jit = {'fit': jit.pop('fit', True),
                  'predict': jit.pop('predict', True),
                  'loss': jit.pop('loss', True)}
      if len(jit):
        raise ValueError(f'Unknown jit setting for {jit.keys()}')
    else:
      raise ValueError(f'Unknown "jit" setting: {jit}')

    # optimizer
    if optimizer is None:
      lr = optim.ExponentialDecay(lr=0.025, decay_steps=1, decay_rate=0.99975)
      optimizer = optim.Adam(lr=lr)
    self.optimizer = optimizer

    # loss
    if isinstance(loss, str):
      loss = getattr(losses, loss)
    elif callable(loss):
      loss = loss
    else:
      raise UnsupportedError(f'Do not support {type(loss)} to specify the loss function. '
                             f'We only support str and callable function.')
    self.loss_fun = loss
    self._train_losses = None
    self._test_losses = None
    self._f_shuffle = None

    # target/output mapping types
    self._mapping_type = None

    # functions
    self._f_loss = dict()
    self._f_train = dict()
    self._f_grad = dict()

    # training parameters
    self.max_grad_norm = max_grad_norm  # gradient clipping
    self.shuffle_data = shuffle_data

    # initialize the optimizer
    if not self.target.is_initialized:
      raise ValueError('Please initialize the target model first by calling "initialize()" function.')
    self.optimizer.register_vars(self.target.vars().subset(bm.TrainVar).unique())

  def __repr__(self):
    name = self.__class__.__name__
    prefix = ' ' * len(name)
    return (f'{name}(target={self.target}, \n\t'
            f'{prefix}jit={self.jit}, \n\t'
            f'{prefix}loss={self.loss_fun}, \n\t'
            f'{prefix}optimizer={self.optimizer})')

  def predict(
      self,
      xs: Union[Tensor, Dict[str, Tensor]],
      forced_states: Dict[str, Tensor] = None,
      forced_feedbacks: Dict[str, Tensor] = None,
      initial_states: Union[Tensor, Dict[str, Tensor]] = None,
      initial_feedbacks: Dict[str, Tensor] = None,
      reset: bool = True,
      shared_kwargs: Dict = None,
      **kwargs
  ):
    """Predict a series of input data with the given target model.

    This function use the JIT compilation to accelerate the model simulation.
    Moreover, it can automatically monitor the node variables, states, inputs,
    feedbacks and its output, if users want.

    Parameters
    ----------
    xs: Tensor, dict
      The feedforward input data. It must be a 3-dimensional data
      which has the shape of `(num_sample, num_time, num_feature)`.
    shared_kwargs: dict
      Shared keyword arguments for the given target model.
    reset: bool
      Whether reset the model states. Default True.

    forced_states: dict
      The fixed node states. Similar with ``xs``, each tensor in
      ``forced_states`` must be a tensor with the shape of
      `(num_sample, num_time, num_feature)`. Default None.

      .. versionadded:: 2.1.4

    forced_feedbacks: dict
      The fixed feedback states. Similar with ``xs``, each tensor in
      ``forced_states`` must be a tensor with the shape of
      `(num_sample, num_time, num_feature)`. Default None.

      .. versionadded:: 2.1.4

    initial_states: JaxArray, ndarray, dict
      The initial states. Each tensor in ``initial_states`` must be a
      tensor with the shape of `(num_sample, num_feature)`.

      .. versionadded:: 2.1.4

    initial_feedbacks: dict
      The initial feedbacks for the node in the network model.
      Each tensor in ``initial_feedbacks`` must be a
      tensor with the shape of `(num_sample, num_feature)`.

      .. versionadded:: 2.1.4

    Returns
    -------
    output: Tensor, dict
      The model output.
    """
    # check forced states/feedbacks
    return super(BPTT, self).predict(xs=xs,
                                     forced_states=forced_states,
                                     forced_feedbacks=forced_feedbacks,
                                     initial_states=initial_states,
                                     initial_feedbacks=initial_feedbacks,
                                     reset=reset,
                                     shared_kwargs=shared_kwargs)

  def fit(
      self,
      train_data: Union[Callable, Sequence],
      test_data: Union[Callable, Sequence] = None,
      num_batch: int = 32,
      num_train: int = 100,
      num_report: int = 100,
      reset: bool = True,
      shared_kwargs: Dict = None,
      forced_states: Dict[str, Tensor] = None,
      forced_feedbacks: Dict[str, Tensor] = None,
      initial_states: Union[Tensor, Dict[str, Tensor]] = None,
      initial_feedbacks: Dict[str, Tensor] = None,
  ):
    """
    Fit the target model according to the given training and testing data.

    Parameters
    ----------
    train_data: callable, sequence of data
      It can be a callable function, or a tuple/list representing `(X, Y)` data.
      - Callable. This function should return a pair of `(X, Y)` data
      - Sequence. It should be a pair of `(X, Y)` train set.
        - ``X``: should be a tensor or a dict of tensors with the shape of
          `(num_sample, num_time, num_feature)`, where `num_sample` is
          the number of samples, `num_time` is the number of the time step,
          and `num_feature` is the number of features.
        - ``Y``: Target values. A tensor or a dict of tensors.
          - If the shape of each tensor is `(num_sample, num_feature)`,
            then we will only fit the model with the only last output.
          - If the shape of each tensor is `(num_sample, num_time, num_feature)`,
            then the fitting happens on the whole data series.
    test_data: callable, sequence of data
      Same as the ``train_data``. It can be a callable function,
      or a tuple/list representing `(X, Y)` data.
    num_batch: int
      The batch size. Default 32. This setting is used when users provide
      the ``train_data`` and ``test_data`` as a pair of `(X, Y)` data, rather
      than a function.
    num_train: int
      The number of training epoch. Default 100.
    num_report: int
      The number of step to report the progress. Default 100 training steps.
    reset: bool
      Whether reset the initial states of the target model.
    shared_kwargs: dict
      The shared keyword arguments for the target models.
    forced_states: dict
      The fixed node states. Similar with ``xs``, each tensor in
      ``forced_states`` must be a tensor with the shape of
      `(num_sample, num_time, num_feature)`.

      .. versionadded:: 2.1.4

    forced_feedbacks: dict
      The fixed feedback states. Similar with ``xs``, each tensor in
      ``forced_states`` must be a tensor with the shape of
      `(num_sample, num_time, num_feature)`.

      .. versionadded:: 2.1.4

    initial_states: JaxArray, ndarray, dict
      The initial states. Each tensor in ``initial_states`` must be a
      tensor with the shape of `(num_sample, num_feature)`.

      .. versionadded:: 2.1.4

    initial_feedbacks: dict
      The initial feedbacks for the node in the network model.
      Each tensor in ``initial_feedbacks`` must be a
      tensor with the shape of `(num_sample, num_feature)`.

      .. versionadded:: 2.1.4

    """
    # training the model
    all_train_losses = []
    all_test_losses = []
    train_i = 0
    t0 = time.time()
    for _ in range(num_train):
      train_data_ = self._get_train_data(train_data, num_batch)

      # training set
      for x, y in train_data_:
        self._set_initial_states(initial_states)
        self._set_initial_feedbacks(initial_feedbacks)
        batch_size = check_data_batch_size(x)
        if reset:
          self.target.initialize(batch_size)
        loss = self.f_train(shared_kwargs)(x, y,
                                           forced_states=forced_states,
                                           forced_feedbacks=forced_feedbacks)
        all_train_losses.append(loss)
        train_i += 1
        if train_i % num_report == 0:
          t1 = time.time()
          print(f'Train {train_i} steps, use {t1 - t0:.4f} s, train loss {round(float(loss), 5)}')
          t0 = t1

      # testing set
      test_data_ = self._get_test_data(test_data, num_batch)
      if test_data_ is not None:
        for x, y in test_data_:
          batch_size = check_data_batch_size(x)
          if reset:
            self.target.initialize(batch_size)
          loss = self.f_loss(shared_kwargs)(x, y,
                                            forced_states=forced_states,
                                            forced_feedbacks=forced_feedbacks)
          all_test_losses.append(loss)

    self._train_losses = bm.asarray(all_train_losses)
    self._test_losses = bm.asarray(all_test_losses)

  def f_grad(self, shared_kwargs=None) -> Callable:
    """Get gradient function."""
    shared_kwargs_str = serialize_kwargs(shared_kwargs)
    if shared_kwargs_str not in self._f_grad:
      self._f_grad[shared_kwargs_str] = self._make_f_grad(shared_kwargs)
    return self._f_grad[shared_kwargs_str]

  def f_loss(self, shared_kwargs=None) -> Callable:
    """Get loss function."""
    shared_kwargs_str = serialize_kwargs(shared_kwargs)
    if shared_kwargs_str not in self._f_loss:
      self._f_loss[shared_kwargs_str] = self._make_f_loss(shared_kwargs)
      if self.jit['loss']:
        dyn_vars = self.target.vars()
        dyn_vars.update(self.dyn_vars)
        self._f_loss[shared_kwargs_str] = bm.jit(self._f_loss[shared_kwargs_str],
                                                 dyn_vars=dyn_vars)
    return self._f_loss[shared_kwargs_str]

  def f_train(self, shared_kwargs=None) -> Callable:
    """Get training function."""
    shared_kwargs_str = serialize_kwargs(shared_kwargs)
    if shared_kwargs_str not in self._f_train:
      self._f_train[shared_kwargs_str] = self._make_f_train(shared_kwargs)
    return self._f_train[shared_kwargs_str]

  @property
  def train_losses(self):
    """Training loss."""
    return self._train_losses

  @property
  def mapping_type(self):
    """Mapping type for the output and the target."""
    return self._mapping_type

  def _make_f_loss(self, shared_kwargs: Dict = None):
    if shared_kwargs is None: shared_kwargs = dict()
    if not isinstance(shared_kwargs, dict):
      raise ValueError(f'Only supports dict for "shared_kwargs". '
                       f'But got {type(shared_kwargs)}: {shared_kwargs}')

    def loss_fun(inputs, targets, forced_states=None, forced_feedbacks=None):
      inputs = self._format_xs(inputs)
      targets = self._format_ys(targets)
      num_batch, num_step = list(inputs.values())[0].shape[:2]
      forced_states = self._check_forced_states(forced_states, num_batch, num_step)
      forced_feedbacks = self._check_forced_feedbacks(forced_feedbacks, num_batch, num_step)
      inputs = {k: bm.moveaxis(v, 0, 1) for k, v in inputs.items()}
      outputs, _ = self._predict(xs=inputs,
                                 shared_kwargs=shared_kwargs,
                                 forced_states=forced_states,
                                 forced_feedbacks=forced_feedbacks)
      outputs = self._format_ys(outputs)
      loss = 0.
      for key, output in outputs.items():
        loss += self.loss_fun(output, targets[key])
      return loss

    return loss_fun

  def _make_f_grad(self, shared_kwargs: Dict = None):
    _f_loss_internal = self._make_f_loss(shared_kwargs)
    dyn_vars = self.target.vars()
    dyn_vars.update(self.dyn_vars)
    tran_vars = dyn_vars.subset(bm.TrainVar)
    return bm.grad(_f_loss_internal,
                   dyn_vars=dyn_vars.unique(),
                   grad_vars=tran_vars.unique(),
                   return_value=True)

  def _make_f_train(self, shared_kwargs: Dict = None):
    if shared_kwargs is None:
      shared_kwargs = dict()
    elif not isinstance(shared_kwargs, dict):
      raise ValueError(f'Only supports dict for "shared_kwargs". '
                       f'But got {type(shared_kwargs)}: {shared_kwargs}')

    def train_func(inputs, targets, forced_states=None, forced_feedbacks=None):
      inputs = self._format_xs(inputs)
      targets = self._format_ys(targets)
      grads, loss = self.f_grad(shared_kwargs)(inputs,
                                               targets,
                                               forced_states=forced_states,
                                               forced_feedbacks=forced_feedbacks)
      if self.max_grad_norm is not None:
        check_float(self.max_grad_norm, 'max_grad_norm', min_bound=0.)
        grads = bm.clip_by_norm(grads, self.max_grad_norm)
      self.optimizer.update(grads)
      return loss

    if self.jit['fit']:
      dyn_vars = self.target.vars()
      dyn_vars.update(self.dyn_vars)
      dyn_vars.update(self.optimizer.vars())
      train_func = bm.jit(train_func, dyn_vars=dyn_vars.unique())
    return train_func

  def _format_ys(self, ys):
    if isinstance(ys, (bm.ndarray, jnp.ndarray)):
      if isinstance(self.target, Network):
        if len(self.target.exit_nodes) != 1:
          raise ValueError(f'The network {self.target} has '
                           f'{len(self.target.exit_nodes)} '
                           f'output nodes, while we only got '
                           f'one output data.')
        ys = {self.target.exit_nodes[0].name: ys}
      else:
        ys = {self.target.name: ys}
    else:
      exit_nodes = self.target.exit_nodes if isinstance(self.target, Network) else [self.target]
      for node in exit_nodes:
        if node.name not in ys:
          raise ValueError(f'The network has output node {node.name}, '
                           f'however, we did not get the corresponding '
                           f'output targets.')
    check_dict_data(ys, key_type=str, val_type=(bm.ndarray, jnp.ndarray))
    return ys

  def _get_train_data(self, train_data, num_batch):
    # training dataset
    if callable(train_data):
      train_data = self._get_data_by_method1(train_data, num_batch)
    elif isinstance(train_data, (tuple, list)):
      if len(train_data) != 2:
        raise ValueError(f"Must be (X, Y) pair, but got a sequence with "
                         f"length {len(train_data)}")
      train_data = self._get_data_by_method2(train_data,
                                             num_batch=num_batch,
                                             shuffle=self.shuffle_data)
    else:
      raise ValueError(f'Train data does not support {type(train_data)}. ')
    return train_data

  def _get_test_data(self, test_data, num_batch):
    # testing dataset
    if test_data is None:
      test_data = None
    elif callable(test_data):
      test_data = self._get_data_by_method1(test_data, num_batch)
    elif isinstance(test_data, (tuple, list)):
      assert len(test_data) == 2, f"Must be (X, Y) pair, but got a sequence with length {len(test_data)}"
      test_data = self._get_data_by_method2(test_data,
                                            num_batch=num_batch,
                                            shuffle=False)
    else:
      raise ValueError(f'Test data does not support {type(test_data)}. ')
    return test_data

  def _get_data_by_method1(self, dataset, num_batch):
    for xs, ys in dataset():
      xs = self._format_xs(xs)
      ys = self._format_ys(ys)
      yield xs, ys

  def _shuffle(self, xs, ys):
    key = jr.PRNGKey(seed=np.random.randint(0, 100000))
    if self._f_shuffle is None:
      def shuffle(xs, ys, key):
        xs = tree_map(lambda x: jr.permutation(key, x, axis=0), xs)
        ys = tree_map(lambda y: jr.permutation(key, y, axis=0), ys)
        return xs, ys

      self._f_shuffle = jit(shuffle)
    return self._f_shuffle(xs, ys, key)

  def _get_data_by_method2(self, dataset, num_batch, shuffle=False, ):
    assert isinstance(dataset, (tuple, list)) and len(dataset) == 2
    xs, ys = dataset
    xs = self._format_xs(xs)
    num_sample = self._get_xs_info(xs)
    ys = self._format_ys(ys)
    if shuffle:
      xs, ys = self._shuffle(xs, ys)

    for data_idx in range(0, num_sample, num_batch):
      if (data_idx + num_batch) > num_sample:
        inputs = {k: v[data_idx:] for k, v in xs.items()}
        targets = {k: v[data_idx:] for k, v in ys.items()}
      else:
        inputs = {k: v[data_idx: data_idx + num_batch] for k, v in xs.items()}
        targets = {k: v[data_idx: data_idx + num_batch] for k, v in ys.items()}
      yield inputs, targets

  def _get_xs_info(self, xs):
    input_shapes = {}
    if isinstance(self.target, Network):
      for node in self.target.entry_nodes:
        name = self.target.entry_nodes[0].name
        input_shapes[name] = node._feedforward_shapes[name]
    else:
      name = self.target.name
      input_shapes[name] = self.target._feedforward_shapes[name]
    num_batch_sizes = []
    for key, val in xs.items():
      if key not in input_shapes:
        raise ValueError(f'Cannot find {key} in the required inputs. Please check!')
      shape = input_shapes[key]
      if bm.ndim(val) != len(shape) + 1:
        raise ValueError(f'Each tensor in "xs" must be a tensor of shape '
                         f'(num_sample, num_time, {str(shape[1:])[1:-1]}). '
                         f'But we got {val.shape}.')
      num_batch_sizes.append(val.shape[0])
    if len(set(num_batch_sizes)) != 1:
      raise ValueError(f'Number of batch size is different across tensors in '
                       f'the provided "xs". We got {set(num_batch_sizes)}.')
    return num_batch_sizes[0]


class BPFF(BPTT):
  """
  The trainer implementing back propagation algorithm
  for feedforward neural networks.

  """

  def __init__(
      self, target: Node, **kwargs
  ):
    super(BPFF, self).__init__(target=target, **kwargs)

  def predict(
      self,
      xs: Union[Tensor, Dict[str, Tensor]],
      initial_states: Union[Tensor, Dict[str, Tensor]] = None,
      initial_feedbacks: Dict[str, Tensor] = None,
      reset: bool = True,
      shared_kwargs: Dict = None,
      forced_states: Dict[str, Tensor] = None,
      forced_feedbacks: Dict[str, Tensor] = None,
      **kwargs
  ):
    """Predict a series of input data with the given target model.

    This function use the JIT compilation to accelerate the model simulation.
    Moreover, it can automatically monitor the node variables, states, inputs,
    feedbacks and its output.

    Parameters
    ----------
    xs: Tensor, dict
      The feedforward input data. It must be a 3-dimensional data
      which has the shape of `(num_sample, num_time, num_feature)`.
    forced_states: None
      The fixed node states.
    forced_feedbacks: None
      The fixed feedback states.
    initial_states: JaxArray, ndarray, dict
      The initial states. Each tensor in ``initial_states`` must be a
      tensor with the shape of `(num_sample, num_feature)`.
    initial_feedbacks: dict
      The initial feedbacks for the node in the network model.
      Each tensor in ``initial_feedbacks`` must be a
      tensor with the shape of `(num_sample, num_feature)`.
    reset: bool
      Whether reset the model states.
    shared_kwargs: optional, dict
      The shared arguments across different layers.

    Returns
    -------
    output: Tensor, dict
      The model output.
    """
    # format input data
    xs = self._format_ys(xs)
    num_batch = self._get_xs_info(xs)
    # get forced data
    forced_states = self._check_forced_states(forced_states, num_batch)
    forced_feedbacks = self._check_forced_feedbacks(forced_feedbacks, num_batch)
    # set initial states
    self._set_initial_states(initial_states)
    self._set_initial_feedbacks(initial_feedbacks)
    # reset the model states
    if reset:
      self.target.initialize(num_batch)
    # init monitor
    for key in self.mon.item_contents.keys():
      self.mon.item_contents[key] = []  # reshape the monitor items
    # prediction
    outputs, hists = self._predict(xs=xs,
                                   forced_states=forced_states,
                                   forced_feedbacks=forced_feedbacks,
                                   shared_kwargs=shared_kwargs)
    # post-running for monitors
    for key in self.mon.item_names:
      self.mon.item_contents[key] = hists[key]
    if self.numpy_mon_after_run:
      self.mon.numpy()
    return outputs

  def _check_forced_states(self, forced_states, num_batch):
    iter_forced_states = dict()
    if forced_states is not None:
      if isinstance(self.target, Network):
        nodes = [node.name for node in self.target.lnodes]
        if not isinstance(forced_states, dict):
          raise ValueError('"forced_states" must be a dict of (str, Tensor)')
        for key, tensor in forced_states.items():
          if not isinstance(key, str):
            raise ValueError(f'"forced_states" must be a dict of (str, tensor). '
                             f'But got a dict of ({type(key)}, {type(tensor)})')
          if key not in nodes:
            raise ValueError(f'Node "{key}" is not defined in the target model. '
                             f'We only detect: \n{self.target.lnodes}')
          if not isinstance(tensor, (bm.ndarray, jnp.ndarray)):
            raise ValueError(f'"forced_states" must a dict of (str, tensor), '
                             f'while we got ({type(key)}, {type(tensor)})')
          if bm.ndim(tensor) != self.target[key].state.ndim:
            raise ValueError(f'Must be a tensor with shape of (num_batch, '
                             f'{str(self.target[key].state.shape)[1:-1]}), '
                             f'but we got {tensor.shape}')
          if tensor.shape[0] != num_batch:
            raise ValueError(f'The number of the batch size ({tensor.shape[0]}) '
                             f'of the forced state of {key} does not '
                             f'match with the batch size in inputs {num_batch}.')
          if self.target[key].output_shape[1:] != tensor.shape[2:]:
            raise UnsupportedError(f'The forced state of {key} has the shape of '
                                   f'{tensor.shape}, which is not consistent with '
                                   f'its output shape {self.target[key].output_shape}. '
                                   f'Each tensor in forced state should have the shape '
                                   f'of (num_sample, num_time, num_feature) or '
                                   f'(num_sample, num_feature).')
          iter_forced_states[key] = bm.moveaxis(tensor, 0, 1)  # shape of (num_time, num_sample, num_feature)
      else:
        raise UnsupportedError('We do not support forced feedback state '
                               'for a single brainpy.nn.Node instance')
    return iter_forced_states

  def _check_forced_feedbacks(self, forced_feedbacks, num_batch):
    iter_forced_feedbacks = dict()
    if forced_feedbacks is not None:
      if isinstance(self.target, Network):
        if not isinstance(forced_feedbacks, dict):
          raise ValueError('"forced_feedbacks" must be a dict of (str, Tensor)')
        feedback_node_names = [node.name for node in self.target.feedback_nodes]
        for key, tensor in forced_feedbacks.items():
          if not isinstance(key, str):
            raise ValueError(f'"forced_feedbacks" must be a dict of (str, tensor). '
                             f'But got a dict of ({type(key)}, {type(tensor)})')
          if key not in feedback_node_names:
            raise ValueError(f'{self.target} has no feedback node {key}, '
                             f'it only has {feedback_node_names}')
          if not isinstance(tensor, (bm.ndarray, jnp.ndarray)):
            raise ValueError('"forced_feedbacks" must a dict of (str, tensor), '
                             'while we got ({type(key)}, {type(tensor)})')
          if bm.ndim(tensor) != self.target[key].fb_output.ndim:
            raise ValueError(f'Must be a tensor with shape of (num_batch, '
                             f'{str(self.target[key].fb_output.shape)[1:-1]}), '
                             f'but we got {tensor.shape}')
          if tensor.shape[0] != num_batch:
            raise ValueError(f'The number of the batch size ({tensor.shape[0]}) '
                             f'of the forced feedback of {key} does not '
                             f'match with the batch size in inputs {num_batch}.')
          if self.target[key].output_shape[1:] != tensor.shape[2:]:
            raise UnsupportedError(f'The forced feedback of {key} has the shape of '
                                   f'{tensor.shape}, which is not consistent with '
                                   f'its output shape {self.target[key].output_shape}. '
                                   f'Each tensor in forced feedback should have the shape '
                                   f'of (num_sample, num_time, num_feature) or '
                                   f'(num_sample, num_feature).')
          iter_forced_feedbacks[key] = bm.moveaxis(tensor, 0, 1)  # shape of (num_time, num_sample, num_feature)
      else:
        raise UnsupportedError('We do not support forced states for '
                               'a single brainpy.nn.Node instance')
    return iter_forced_feedbacks

  def _predict(
      self,
      xs: Dict[str, Tensor],
      shared_kwargs: Dict = None,
      forced_states: Dict[str, Tensor] = None,
      forced_feedbacks: Dict[str, Tensor] = None,
  ):
    """Predict the output according to the inputs.

    Parameters
    ----------
    xs: dict
      Each tensor should have the shape of `(num_time, num_batch, num_feature)`.
    forced_states: dict
      The forced state values.
    forced_feedbacks: dict
      The forced feedback output values.
    shared_kwargs: optional, dict
      The shared keyword arguments.

    Returns
    -------
    outputs, hists
      A tuple of pair of (outputs, hists).
    """
    _predict_func = self._get_predict_func(shared_kwargs)
    # rune the model
    forced_states = dict() if forced_states is None else forced_states
    forced_feedbacks = dict() if forced_feedbacks is None else forced_feedbacks
    return _predict_func(xs, forced_states, forced_feedbacks)

  def _make_f_loss(self, shared_kwargs: Dict = None):
    if shared_kwargs is None: shared_kwargs = dict()
    if not isinstance(shared_kwargs, dict):
      raise ValueError(f'Only supports dict for "shared_kwargs". '
                       f'But got {type(shared_kwargs)}: {shared_kwargs}')

    def loss_fun(inputs, targets, forced_states=None, forced_feedbacks=None):
      inputs = self._format_xs(inputs)
      targets = self._format_ys(targets)
      num_batch, num_step = list(inputs.values())[0].shape[:2]
      forced_states = self._check_forced_states(forced_states, num_batch)
      forced_feedbacks = self._check_forced_feedbacks(forced_feedbacks, num_batch)
      outputs, _ = self._predict(xs=inputs,
                                 shared_kwargs=shared_kwargs,
                                 forced_states=forced_states,
                                 forced_feedbacks=forced_feedbacks)
      outputs = self._format_ys(outputs)
      loss = 0.
      for key, output in outputs.items():
        loss += self.loss_fun(output, targets[key])
      return loss

    return loss_fun

  def _get_predict_func(self, shared_kwargs: Dict = None):
    if shared_kwargs is None: shared_kwargs = dict()
    shared_kwargs_str = serialize_kwargs(shared_kwargs)
    if shared_kwargs_str not in self._predict_func:
      self._predict_func[shared_kwargs_str] = self._make_predict_func(shared_kwargs)
    return self._predict_func[shared_kwargs_str]

  def _make_predict_func(self, shared_kwargs: Dict):
    if not isinstance(shared_kwargs, dict):
      raise ValueError(f'"shared_kwargs" must be a dict, '
                       f'but got {type(shared_kwargs)}')

    def run_func(xs, forced_states, forced_feedbacks):
      monitors = self.mon.item_contents.keys()
      return self.target(xs,
                         forced_states=forced_states,
                         forced_feedbacks=forced_feedbacks,
                         monitors=monitors,
                         **shared_kwargs)

    if self.jit['predict']:
      dyn_vars = self.target.vars()
      dyn_vars.update(self.dyn_vars)
      run_func = bm.jit(run_func, dyn_vars=dyn_vars.unique())
    return run_func

  def _get_xs_info(self, xs):
    input_shapes = {}
    if isinstance(self.target, Network):
      for node in self.target.entry_nodes:
        name = self.target.entry_nodes[0].name
        input_shapes[name] = node._feedforward_shapes[name]
    else:
      name = self.target.name
      input_shapes[name] = self.target._feedforward_shapes[name]
    num_batch_sizes = []
    for key, val in xs.items():
      if key not in input_shapes:
        raise ValueError(f'Cannot find {key} in the required inputs. Please check!')
      shape = input_shapes[key]
      if bm.ndim(val) != len(shape):
        raise ValueError(f'Each tensor in "xs" must be a tensor of shape '
                         f'(num_sample, {str(shape[1:])[1:-1]}). '
                         f'But we got {val.shape}.')
      num_batch_sizes.append(val.shape[0])
    if len(set(num_batch_sizes)) != 1:
      raise ValueError(f'Number of batch size is different across tensors in '
                       f'the provided "xs". We got {set(num_batch_sizes)}.')
    return num_batch_sizes[0]



# -*- coding: utf-8 -*-

from typing import Dict, Callable

from brainpy import math as bm, errors
from brainpy.integrators import constants, utils
from brainpy.integrators.base import Integrator
from brainpy.integrators.constants import DT

__all__ = [
  'SDEIntegrator',
]


def f_names(f):
  func_name = constants.unique_name('sde')
  if f.__name__.isidentifier():
    func_name += '_' + f.__name__
  return func_name


class SDEIntegrator(Integrator):
  """SDE Integrator."""

  def __init__(
      self,
      f: Callable,
      g: Callable,
      dt: float = None,
      name: str = None,
      show_code: bool = False,
      var_type: str = None,
      intg_type: str = None,
      wiener_type: str = None,
      state_delays: Dict[str, bm.AbstractDelay] = None,
  ):

    dt = bm.get_dt() if dt is None else dt
    parses = utils.get_args(f)
    variables = parses[0]  # variable names, (before 't')
    parameters = parses[1]  # parameter names, (after 't')
    arguments = parses[2]  # function arguments

    # super initialization
    super(SDEIntegrator, self).__init__(name=name,
                                        variables=variables,
                                        parameters=parameters,
                                        arguments=arguments,
                                        dt=dt,
                                        state_delays=state_delays)

    # derivative functions
    self.derivative = {constants.F: f, constants.G: g}
    self.f = f
    self.g = g

    # essential parameters
    intg_type = constants.ITO_SDE if intg_type is None else intg_type
    var_type = constants.SCALAR_VAR if var_type is None else var_type
    wiener_type = constants.SCALAR_WIENER if wiener_type is None else wiener_type
    if intg_type not in constants.SUPPORTED_INTG_TYPE:
      raise errors.IntegratorError(f'Currently, BrainPy only support SDE_INT types: '
                                   f'{constants.SUPPORTED_INTG_TYPE}. But we got {intg_type}.')
    if var_type not in constants.SUPPORTED_VAR_TYPE:
      raise errors.IntegratorError(f'Currently, BrainPy only supports variable types: '
                                   f'{constants.SUPPORTED_VAR_TYPE}. But we got {var_type}.')
    if wiener_type not in constants.SUPPORTED_WIENER_TYPE:
      raise errors.IntegratorError(f'Currently, BrainPy only supports Wiener '
                                   f'Process types: {constants.SUPPORTED_WIENER_TYPE}. '
                                   f'But we got {wiener_type}.')
    self.var_type = var_type  # variable type
    self.intg_type = intg_type  # integral type
    self.wiener_type = wiener_type  # wiener process type

    # random seed
    self.rng = bm.random.RandomState()

    # code scope
    self.code_scope = {constants.F: f, constants.G: g, 'math': bm, 'random': self.rng}

    # code lines
    self.func_name = f_names(f)
    self.code_lines = [f'def {self.func_name}({", ".join(self.arguments)}):']

    # others
    self.show_code = show_code



# -*- coding: utf-8 -*-

import jax.numpy as jnp
import numpy as np

from brainpy import math as bm
from brainpy.tools.checking import check_float, check_integer

__all__ = [
  'section_input',
  'constant_input', 'constant_current',
  'spike_input', 'spike_current',
  'ramp_input', 'ramp_current',
  'wiener_process',
  'ou_process',
  'sinusoidal_input',
  'square_input',
]


def section_input(values, durations, dt=None, return_length=False):
  """Format an input current with different sections.

  For example:

  If you want to get an input where the size is 0 bwteen 0-100 ms,
  and the size is 1. between 100-200 ms.

  >>> section_input(values=[0, 1],
  >>>               durations=[100, 100])

  Parameters
  ----------
  values : list, np.ndarray
      The current values for each period duration.
  durations : list, np.ndarray
      The duration for each period.
  dt : float
      Default is None.
  return_length : bool
      Return the final duration length.

  Returns
  -------
  current_and_duration
  """
  if len(durations) != len(values):
    raise ValueError(f'"values" and "durations" must be the same length, while '
                     f'we got {len(values)} != {len(durations)}.')

  dt = bm.get_dt() if dt is None else dt

  # get input current shape, and duration
  I_duration = sum(durations)
  I_shape = ()
  for val in values:
    shape = jnp.shape(val)
    if len(shape) > len(I_shape):
      I_shape = shape

  # get the current
  start = 0
  I_current = bm.zeros((int(np.ceil(I_duration / dt)),) + I_shape, dtype=bm.float_)
  for c_size, duration in zip(values, durations):
    length = int(duration / dt)
    I_current[start: start + length] = c_size
    start += length

  if return_length:
    return I_current.value, I_duration
  else:
    return I_current.value


def constant_input(I_and_duration, dt=None):
  """Format constant input in durations.

  For example:

  If you want to get an input where the size is 0 bwteen 0-100 ms,
  and the size is 1. between 100-200 ms.

  >>> import brainpy.math as bm
  >>> constant_input([(0, 100), (1, 100)])
  >>> constant_input([(bm.zeros(100), 100), (bm.random.rand(100), 100)])

  Parameters
  ----------
  I_and_duration : list
      This parameter receives the current size and the current
      duration pairs, like `[(Isize1, duration1), (Isize2, duration2)]`.
  dt : float
      Default is None.

  Returns
  -------
  current_and_duration : tuple
      (The formatted current, total duration)
  """
  dt = bm.get_dt() if dt is None else dt

  # get input current dimension, shape, and duration
  I_duration = 0.
  I_shape = ()
  for I in I_and_duration:
    I_duration += I[1]
    shape = jnp.shape(I[0])
    if len(shape) > len(I_shape):
      I_shape = shape

  # get the current
  start = 0
  I_current = jnp.zeros((int(np.ceil(I_duration / dt)),) + I_shape, dtype=bm.float_)
  for c_size, duration in I_and_duration:
    length = int(duration / dt)
    I_current = I_current.at[start: start + length].set(c_size)
    start += length
  return I_current, I_duration


constant_current = constant_input


def spike_input(sp_times, sp_lens, sp_sizes, duration, dt=None):
  """Format current input like a series of short-time spikes.

  For example:

  If you want to generate a spike train at 10 ms, 20 ms, 30 ms, 200 ms, 300 ms,
  and each spike lasts 1 ms and the spike current is 0.5, then you can use the
  following funtions:

  >>> spike_input(sp_times=[10, 20, 30, 200, 300],
  >>>             sp_lens=1.,  # can be a list to specify the spike length at each point
  >>>             sp_sizes=0.5,  # can be a list to specify the current size at each point
  >>>             duration=400.)

  Parameters
  ----------
  sp_times : list, tuple
      The spike time-points. Must be an iterable object.
  sp_lens : int, float, list, tuple
      The length of each point-current, mimicking the spike durations.
  sp_sizes : int, float, list, tuple
      The current sizes.
  duration : int, float
      The total current duration.
  dt : float
      The default is None.

  Returns
  -------
  current : bm.ndarray
      The formatted input current.
  """
  dt = bm.get_dt() if dt is None else dt
  assert isinstance(sp_times, (list, tuple))
  if isinstance(sp_lens, (float, int)):
    sp_lens = [sp_lens] * len(sp_times)
  if isinstance(sp_sizes, (float, int)):
    sp_sizes = [sp_sizes] * len(sp_times)

  current = jnp.zeros(int(np.ceil(duration / dt)), dtype=bm.float_)
  for time, dur, size in zip(sp_times, sp_lens, sp_sizes):
    pp = int(time / dt)
    p_len = int(dur / dt)
    current = current.at[pp: pp + p_len].set(size)
  return current


spike_current = spike_input


def ramp_input(c_start, c_end, duration, t_start=0, t_end=None, dt=None):
  """Get the gradually changed input current.

  Parameters
  ----------
  c_start : float
      The minimum (or maximum) current size.
  c_end : float
      The maximum (or minimum) current size.
  duration : int, float
      The total duration.
  t_start : float
      The ramped current start time-point.
  t_end : float
      The ramped current end time-point. Default is the None.
  dt : float, int, optional
      The numerical precision.

  Returns
  -------
  current : bm.ndarray
    The formatted current
  """
  dt = bm.get_dt() if dt is None else dt
  t_end = duration if t_end is None else t_end

  current = jnp.zeros(int(np.ceil(duration / dt)), dtype=bm.float_)
  p1 = int(np.ceil(t_start / dt))
  p2 = int(np.ceil(t_end / dt))
  cc = jnp.array(jnp.linspace(c_start, c_end, p2 - p1), dtype=bm.float_)
  current = current.at[p1: p2].set(cc)
  return current


ramp_current = ramp_input


def wiener_process(duration, dt=None, n=1, t_start=0., t_end=None, seed=None):
  """Stimulus sampled from a Wiener process, i.e.
  drawn from standard normal distribution N(0, sqrt(dt)).

  Parameters
  ----------
  duration: float
    The input duration.
  dt: float
    The numerical precision.
  n: int
    The variable number.
  t_start: float
    The start time.
  t_end: float
    The end time.
  seed: int
    The noise seed.
  """
  dt = bm.get_dt() if dt is None else dt
  check_float(dt, 'dt', allow_none=False, min_bound=0.)
  check_integer(n, 'n', allow_none=False, min_bound=0)
  rng = bm.random.RandomState(seed)
  t_end = duration if t_end is None else t_end
  i_start = int(t_start / dt)
  i_end = int(t_end / dt)
  noises = rng.standard_normal((i_end - i_start, n)) * jnp.sqrt(dt)
  currents = jnp.zeros((int(duration / dt), n))
  currents = currents.at[i_start: i_end].set(bm.as_device_array(noises))
  return currents


def ou_process(mean, sigma, tau, duration, dt=None, n=1, t_start=0., t_end=None, seed=None):
  r"""Ornstein–Uhlenbeck input.

  .. math::

     dX = (mu - X)/\tau * dt + \sigma*dW

  Parameters
  ----------
  mean: float
    Drift of the OU process.
  sigma: float
    Standard deviation of the Wiener process, i.e. strength of the noise.
  tau: float
    Timescale of the OU process, in ms.
  duration: float
    The input duration.
  dt: float
    The numerical precision.
  n: int
    The variable number.
  t_start: float
    The start time.
  t_end: float
    The end time.

  """
  dt = bm.get_dt() if dt is None else dt
  dt_sqrt = jnp.sqrt(dt)
  check_float(dt, 'dt', allow_none=False, min_bound=0.)
  check_integer(n, 'n', allow_none=False, min_bound=0)
  rng = bm.random.RandomState(seed)
  x = bm.Variable(jnp.ones(n) * mean)

  def _f(t):
    x.value = x + dt * ((mean - x) / tau) + sigma * dt_sqrt * rng.standard_normal(n)

  f = bm.make_loop(_f, dyn_vars=[x, rng], out_vars=x)
  noises = f(jnp.arange(t_start, t_end, dt))

  t_end = duration if t_end is None else t_end
  i_start = int(t_start / dt)
  i_end = int(t_end / dt)
  currents = bm.zeros((int(duration / dt), n))
  currents[i_start: i_end] = noises
  return currents.value


def sinusoidal_input(amplitude, frequency, duration, dt=None, t_start=0., t_end=None, dc_bias=False):
  """Sinusoidal input.

  Parameters
  ----------
  amplitude: float
    Amplitude of the sinusoid.
  frequency: float
    Frequency of the sinus oscillation, in Hz
  duration: float
    The input duration.
  t_start: float
    The start time.
  t_end: float
    The end time.
  dt: float
    The numerical precision.
  dc_bias: bool
    Whether the sinusoid oscillates around 0 (False), or
    has a positive DC bias, thus non-negative (True).
  """
  dt = bm.get_dt() if dt is None else dt
  check_float(dt, 'dt', allow_none=False, min_bound=0.)
  if t_end is None:
    t_end = duration
  times = jnp.arange(0, t_end - t_start, dt)
  start_i = int(t_start / dt)
  end_i = int(t_end / dt)
  sin_inputs = amplitude * jnp.sin(2 * jnp.pi * times * (frequency / 1000.0))
  if dc_bias:
    sin_inputs += amplitude
  currents = bm.zeros(int(duration / dt))
  currents[start_i:end_i] = sin_inputs
  return currents.value


def _square(t, duty=0.5):
  t, w = np.asarray(t), np.asarray(duty)
  w = np.asarray(w + (t - t))
  t = np.asarray(t + (w - w))
  if t.dtype.char in ['fFdD']:
    ytype = t.dtype.char
  else:
    ytype = 'd'

  y = np.zeros(t.shape, ytype)

  # width must be between 0 and 1 inclusive
  mask1 = (w > 1) | (w < 0)
  np.place(y, mask1, np.nan)

  # on the interval 0 to duty*2*pi function is 1
  tmod = np.mod(t, 2 * np.pi)
  mask2 = (1 - mask1) & (tmod < w * 2 * np.pi)
  np.place(y, mask2, 1)

  # on the interval duty*2*pi to 2*pi function is
  #  (pi*(w+1)-tmod) / (pi*(1-w))
  mask3 = (1 - mask1) & (1 - mask2)
  np.place(y, mask3, -1)
  return y


def square_input(amplitude, frequency, duration, dt=None, dc_bias=False, t_start=None, t_end=None):
  """Oscillatory square input.

  Parameters
  ----------
  amplitude: float
    Amplitude of the square oscillation.
  frequency: float
    Frequency of the square oscillation, in Hz.
  duration: float
    The input duration.
  t_start: float
    The start time.
  t_end: float
    The end time.
  dt: float
    The numerical precision.
  dc_bias: bool
    Whether the sinusoid oscillates around 0 (False), or
    has a positive DC bias, thus non-negative (True).
  """
  dt = bm.get_dt() if dt is None else dt
  check_float(dt, 'dt', allow_none=False, min_bound=0.)
  if t_end is None:
    t_end = duration
  times = jnp.arange(0, t_end - t_start, dt)
  currents = bm.zeros(int(duration / dt))
  start_i = int(t_start / dt)
  end_i = int(t_end / dt)
  sin_inputs = amplitude * _square(2 * jnp.pi * times * (frequency / 1000.0))
  if dc_bias:
    sin_inputs += amplitude
  currents[start_i:end_i] = sin_inputs
  return currents.value



# -*- coding: utf-8 -*-


__all__ = [
  # data types
  'DataType',

  # pass rules
  'SingleData',
  'MultipleData',
]


class DataType(object):
  """Base class for data type."""

  def filter(self, data):
    raise NotImplementedError

  def __repr__(self):
    return self.__class__.__name__


class SingleData(DataType):
  """Pass the only one data into the node.
  If there are multiple data, an error will be raised. """

  def filter(self, data):
    if data is None:
      return None
    if len(data) > 1:
      raise ValueError(f'{self.__class__.__name__} only support one '
                       f'feedforward/feedback input. But we got {len(data)}.')
    return tuple(data.values())[0]

  def __repr__(self):
    return self.__class__.__name__


class MultipleData(DataType):
  """Pass a list/tuple of data into the node."""

  def __init__(self, return_type: str = 'sequence'):
    if return_type not in ['sequence', 'name_dict', 'type_dict', 'node_dict']:
      raise ValueError(f"Only support return type of 'sequence', 'name_dict', "
                       f"'type_dict' and 'node_dict'. But we got {return_type}")
    self.return_type = return_type

    from brainpy.nn.base import Node

    if return_type == 'sequence':
      f = lambda data: tuple(data.values())

    elif return_type == 'name_dict':
      # Pass a dict with <node name, data> into the node.

      def f(data):
        _res = dict()
        for node, val in data.items():
          if isinstance(node, str):
            _res[node] = val
          elif isinstance(node, Node):
            _res[node.name] = val
          else:
            raise ValueError(f'Unknown type {type(node)}: node')
        return _res

    elif return_type == 'type_dict':
      # Pass a dict with <node type, data> into the node.

      def f(data):
        _res = dict()
        for node, val in data.items():
          if isinstance(node, str):
            _res[str] = val
          elif isinstance(node, Node):
            _res[type(node.name)] = val
          else:
            raise ValueError(f'Unknown type {type(node)}: node')
        return _res

    elif return_type == 'node_dict':
      # Pass a dict with <node, data> into the node.
      f = lambda data: data

    else:
      raise ValueError
    self.return_func = f

  def __repr__(self):
    return f'{self.__class__.__name__}(return_type={self.return_type})'

  def filter(self, data):
    if data is None:
      return None
    else:
      return self.return_func(data)



# -*- coding: utf-8 -*-

import inspect
import re
from types import LambdaType

__all__ = [
  'copy_doc',
  'code_lines_to_func',

  # tools for code string
  'get_identifiers',
  'indent',
  'deindent',
  'word_replace',

  # other tools
  'is_lambda_function',
  'get_main_code',
  'get_func_source',
  'change_func_name',
]


def copy_doc(source_f):
  def copy(target_f):
    target_f.__doc__ = source_f.__doc__
    return target_f

  return copy


def code_lines_to_func(lines, func_name, func_args, scope, remind=''):
  lines_for_compile = [f'    {line}' for line in lines]
  code_for_compile = '\n'.join(lines_for_compile)
  code = f'def {func_name}({", ".join(func_args)}):\n' + \
         f'  try:\n' + \
         f'{code_for_compile}\n' + \
         f'  except Exception as e:\n'
  code += '    exc_type, exc_obj, exc_tb = sys.exc_info()\n'
  code += '    line_no = exc_tb.tb_lineno\n'
  code += '    raise ValueError(f"Error occurred in line {line_no}: {code_for_debug} {str(e)} {remind}")'
  lines_for_debug = [f'[{i + 1:3d}] {line}' for i, line in enumerate(code.split('\n'))]
  code_for_debug = '\n'.join(lines_for_debug)
  scope['code_for_debug'] = '\n\n' + code_for_debug + '\n\n'
  scope['remind'] = '\n' + remind + '\n'
  try:
    exec(compile(code, '', 'exec'), scope)
  except Exception as e:
    raise ValueError(f'Compilation function error: \n\n{code}') from e
  func = scope[func_name]
  return code, func


######################################
# String tools
######################################


def get_identifiers(expr, include_numbers=False):
  """
  Return all the identifiers in a given string ``expr``, that is everything
  that matches a programming language variable like expression, which is
  here implemented as the regexp ``\\b[A-Za-z_][A-Za-z0-9_]*\\b``.

  Parameters
  ----------
  expr : str
      The string to analyze
  include_numbers : bool, optional
      Whether to include number literals in the output. Defaults to ``False``.

  Returns
  -------
  identifiers : set
      A set of all the identifiers (and, optionally, numbers) in `expr`.

  Examples
  --------
  >>> expr = '3-a*_b+c5+8+f(A - .3e-10, tau_2)*17'
  >>> ids = get_identifiers(expr)
  >>> print(sorted(list(ids)))
  ['A', '_b', 'a', 'c5', 'f', 'tau_2']
  >>> ids = get_identifiers(expr, include_numbers=True)
  >>> print(sorted(list(ids)))
  ['.3e-10', '17', '3', '8', 'A', '_b', 'a', 'c5', 'f', 'tau_2']
  """

  _ID_KEYWORDS = {'and', 'or', 'not', 'True', 'False'}
  identifiers = set(re.findall(r'\b[A-Za-z_][A-Za-z0-9_.]*\b', expr))
  # identifiers = set(re.findall(r'\b[A-Za-z_][.?[A-Za-z0-9_]*]*\b', expr))
  if include_numbers:
    # only the number, not a + or -
    pattern = r'(?<=[^A-Za-z_])[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?|^[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?'
    numbers = set(re.findall(pattern, expr))
  else:
    numbers = set()
  return (identifiers - _ID_KEYWORDS) | numbers


def indent(text, num_tabs=1, spaces_per_tab=4, tab=None):
  if tab is None:
    tab = ' ' * spaces_per_tab
  indent_ = tab * num_tabs
  indented_string = indent_ + text.replace('\n', '\n' + indent_)
  return indented_string


def deindent(text, num_tabs=None, spaces_per_tab=4, docstring=False):
  text = text.replace('\t', ' ' * spaces_per_tab)
  lines = text.split('\n')
  # if it's a docstring, we search for the common tabulation starting from
  # line 1, otherwise we use all lines
  if docstring:
    start = 1
  else:
    start = 0
  if docstring and len(lines) < 2:  # nothing to do
    return text
  # Find the minimum indentation level
  if num_tabs is not None:
    indent_level = num_tabs * spaces_per_tab
  else:
    line_seq = [len(line) - len(line.lstrip()) for line in lines[start:] if len(line.strip())]
    if len(line_seq) == 0:
      indent_level = 0
    else:
      indent_level = min(line_seq)
  # remove the common indentation
  lines[start:] = [line[indent_level:] for line in lines[start:]]
  return '\n'.join(lines)


def word_replace(expr, substitutions, exclude_dot=True):
  """Applies a dict of word substitutions.

  The dict ``substitutions`` consists of pairs ``(word, rep)`` where each
  word ``word`` appearing in ``expr`` is replaced by ``rep``. Here a 'word'
  means anything matching the regexp ``\\bword\\b``.

  Examples
  --------

  >>> expr = 'a*_b+c5+8+f(A)'
  >>> print(word_replace(expr, {'a':'banana', 'f':'func'}))
  banana*_b+c5+8+func(A)
  """
  for var, replace_var in substitutions.items():
    if exclude_dot:
      expr = re.sub(r'\b(?<!\.)' + var + r'\b(?!\.)', str(replace_var), expr)
    else:
      expr = re.sub(r'\b' + var + r'\b', str(replace_var), expr)
  return expr


######################################
# Other tools
######################################


def change_func_name(f, name):
  f.__name__ = name
  return f


def is_lambda_function(func):
  """Check whether the function is a ``lambda`` function. Comes from
  https://stackoverflow.com/questions/23852423/how-to-check-that-variable-is-a-lambda-function

  Parameters
  ----------
  func : callable function
      The function.

  Returns
  -------
  bool
      True of False.
  """
  return isinstance(func, LambdaType) and func.__name__ == "<lambda>"


def get_func_source(func):
  code = inspect.getsource(func)
  # remove @
  try:
    start = code.index('def ')
    code = code[start:]
  except ValueError:
    pass
  return code


def get_main_code(func, codes=None):
  """Get the main function _code string.

  For lambda function, return the

  Parameters
  ----------
  func : callable, Optional, int, float

  Returns
  -------

  """
  if func is None:
    return ''
  elif callable(func):
    if is_lambda_function(func):
      codes = (codes or get_func_source(func))
      splits = codes.split(':')
      if len(splits) != 2:
        raise ValueError(f'Can not parse function: \n{codes}')
      return f'return {splits[1]}'

    else:
      codes = (codes.split('\n') or inspect.getsourcelines(func)[0])
      idx = 0
      for line in codes:
        idx += 1
        line = line.replace(' ', '')
        if '):' in line:
          break
      else:
        code = "\n".join(codes)
        raise ValueError(f'Can not parse function: \n{code}')
      return ''.join(codes[idx:])
  else:
    raise ValueError(f'Unknown function type: {type(func)}.')



# -*- coding: utf-8 -*-

import os
import re

import jax.config
import jax.numpy as jnp

__all__ = [
  'enable_x64',
  'disable_x64',
  'set_platform',
  'set_host_device_count',

  # data types
  'bool_',
  'int_',
  'float_',
  'complex_',

  # change default data types
  'set_dt',
  'get_dt',
]

# default dtype
# --------------------------

bool_ = jnp.bool_
int_ = jnp.int32
float_ = jnp.float32
complex_ = jnp.complex_


# numerical precision
# --------------------------

__dt = 0.1


def set_dt(dt):
  """Set the numerical integrator precision.

  Parameters
  ----------
  dt : float
      Numerical integration precision.
  """
  assert isinstance(dt, float), f'"dt" must a float, but we got {dt}'
  global __dt
  __dt = dt


def get_dt():
  """Get the numerical integrator precision.

  Returns
  -------
  dt : float
      Numerical integration precision.
  """
  return __dt


def enable_x64(mode=True):
  assert mode in [True, False]
  jax.config.update("jax_enable_x64", mode)


def disable_x64():
  jax.config.update("jax_enable_x64", False)


def set_platform(platform):
  """
  Changes platform to CPU, GPU, or TPU. This utility only takes
  effect at the beginning of your program.
  """
  assert platform in ['cpu', 'gpu', 'tpu']
  jax.config.update("jax_platform_name", platform)


def set_host_device_count(n):
  """
  By default, XLA considers all CPU cores as one device. This utility tells XLA
  that there are `n` host (CPU) devices available to use. As a consequence, this
  allows parallel mapping in JAX :func:`jax.pmap` to work in CPU platform.

  .. note:: This utility only takes effect at the beginning of your program.
      Under the hood, this sets the environment variable
      `XLA_FLAGS=--xla_force_host_platform_device_count=[num_devices]`, where
      `[num_device]` is the desired number of CPU devices `n`.

  .. warning:: Our understanding of the side effects of using the
      `xla_force_host_platform_device_count` flag in XLA is incomplete. If you
      observe some strange phenomenon when using this utility, please let us
      know through our issue or forum page. More information is available in this
      `JAX issue <https://github.com/google/jax/issues/1408>`_.

  :param int n: number of devices to use.
  """
  xla_flags = os.getenv("XLA_FLAGS", "")
  xla_flags = re.sub(r"--xla_force_host_platform_device_count=\S+", "", xla_flags).split()
  os.environ["XLA_FLAGS"] = " ".join(["--xla_force_host_platform_device_count={}".format(n)] + xla_flags)



# -*- coding: utf-8 -*-

import multiprocessing

__all__ = [
  'process_pool',
  'process_pool_lock',
]


def process_pool(func, all_net_params, nb_process):
  """Run multiple models in multi-processes.

  Parameters
  ----------
  func : callable
      The function to run model.
  all_net_params : a_list, tuple
      The parameters of the function arguments.
      The parameters for each process can be a tuple, or a dictionary.
  nb_process : int
      The number of the processes.

  Returns
  -------
  results : list
      Process results.
  """
  print('{} jobs total.'.format(len(all_net_params)))
  pool = multiprocessing.Pool(processes=nb_process)
  results = []
  for net_params in all_net_params:
    if isinstance(net_params, (list, tuple)):
      results.append(pool.apply_async(func, args=tuple(net_params)))
    elif isinstance(net_params, dict):
      results.append(pool.apply_async(func, kwds=net_params))
    else:
      raise ValueError('Unknown parameter type: ', type(net_params))
  pool.close()
  pool.join()
  return results


def process_pool_lock(func, all_net_params, nb_process):
  """Run multiple models in multi-processes with lock.

  Sometimes, you want to synchronize the processes. For example,
  if you want to write something in a document, you cannot let
  multi-process simultaneously open this same file. So, you need
  add a `lock` argument in your defined `func`:

  .. code-block:: python

      def some_func(..., lock, ...):
          ... do something ..

          lock.acquire()
          ... something cannot simultaneously do by multi-process ..
          lock.release()

  In such case, you can use `process_pool_lock()` to run your model.

  Parameters
  ----------
  func : callable
      The function to run model.
  all_net_params : a_list, tuple
      The parameters of the function arguments.
  nb_process : int
      The number of the processes.

  Returns
  -------
  results : list
      Process results.
  """
  print('{} jobs total.'.format(len(all_net_params)))
  pool = multiprocessing.Pool(processes=nb_process)
  m = multiprocessing.Manager()
  lock = m.Lock()
  results = []
  for net_params in all_net_params:
    if isinstance(net_params, (list, tuple)):
      results.append(pool.apply_async(func, args=tuple(net_params) + (lock,)))
    elif isinstance(net_params, dict):
      net_params.update(lock=lock)
      results.append(pool.apply_async(func, kwds=net_params))
    else:
      raise ValueError('Unknown parameter type: ', type(net_params))
  pool.close()
  pool.join()
  return results



# -*- coding: utf-8 -*-

import logging

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import animation
from matplotlib.gridspec import GridSpec

from brainpy import math, errors

logger = logging.getLogger('brainpy.visualization')


__all__ = [
  'line_plot',
  'raster_plot',
  'animate_2D',
  'animate_1D',
]


def line_plot(ts,
              val_matrix,
              plot_ids=None,
              ax=None,
              xlim=None,
              ylim=None,
              xlabel='Time (ms)',
              ylabel=None,
              legend=None,
              title=None,
              show=False,
              **kwargs):
  """Show the specified value in the given object (Neurons or Synapses.)

  Parameters
  ----------
  ts : np.ndarray
      The time steps.
  val_matrix : np.ndarray
      The value matrix which record the history trajectory.
      It can be easily accessed by specifying the ``monitors``
      of NeuGroup/SynConn by:
      ``neu/syn = NeuGroup/SynConn(..., monitors=[k1, k2])``
  plot_ids : None, int, tuple, a_list
      The index of the value to plot.
  ax : None, Axes
      The figure to plot.
  xlim : list, tuple
      The xlim.
  ylim : list, tuple
      The ylim.
  xlabel : str
      The xlabel.
  ylabel : str
      The ylabel.
  legend : str
      The prefix of legend for plot.
  show : bool
      Whether show the figure.
  """
  # get plot_ids
  if plot_ids is None:
    plot_ids = [0]
  elif isinstance(plot_ids, int):
    plot_ids = [plot_ids]
  if not isinstance(plot_ids, (list, tuple)) and \
      not (isinstance(plot_ids, np.ndarray) and np.ndim(plot_ids) == 1):
    raise errors.BrainPyError(f'"plot_ids" specifies the value index to plot, it must '
                              f'be a list/tuple/1D numpy.ndarray, not {type(plot_ids)}.')

  # get ax
  if ax is None:
    ax = plt

  val_matrix = val_matrix.reshape((val_matrix.shape[0], -1))
  # change data
  val_matrix = np.asarray(val_matrix)
  ts = np.asarray(ts)

  # plot
  if legend:
    for idx in plot_ids:
      label = legend if len(plot_ids) == 1 else f'{legend}-{idx}'
      ax.plot(ts, val_matrix[:, idx], label=label, **kwargs)
  else:
    for idx in plot_ids:
      ax.plot(ts, val_matrix[:, idx], **kwargs)

  # legend
  if legend:
    ax.legend()

  # xlim
  if xlim is not None:
    plt.xlim(xlim[0], xlim[1])

  # ylim
  if ylim is not None:
    plt.ylim(ylim[0], ylim[1])

  # xlable
  if xlabel:
    plt.xlabel(xlabel)

  # ylabel
  if ylabel:
    plt.ylabel(ylabel)

  # title
  if title:
    plt.title(title)

  # show
  if show:
    plt.show()


def raster_plot(ts,
                sp_matrix,
                ax=None,
                marker='.',
                markersize=2,
                color='k',
                xlabel='Time (ms)',
                ylabel='Neuron index',
                xlim=None,
                ylim=None,
                title=None,
                show=False,
                **kwargs):
  """Show the rater plot of the spikes.

  Parameters
  ----------
  ts : np.ndarray
      The run times.
  sp_matrix : np.ndarray
      The spike matrix which records the spike information.
      It can be easily accessed by specifying the ``monitors``
      of NeuGroup by: ``neu = NeuGroup(..., monitors=['spike'])``
  ax : Axes
      The figure.
  markersize : int
      The size of the marker.
  color : str
      The color of the marker.
  xlim : list, tuple
      The xlim.
  ylim : list, tuple
      The ylim.
  xlabel : str
      The xlabel.
  ylabel : str
      The ylabel.
  show : bool
      Show the figure.
  """

  sp_matrix = np.asarray(sp_matrix)
  if ts is None:
    raise errors.BrainPyError('Must provide "ts".')
  ts = np.asarray(ts)

  # get index and time
  elements = np.where(sp_matrix > 0.)
  index = elements[1]
  time = ts[elements[0]]

  # plot rater
  if ax is None:
    ax = plt
  ax.plot(time, index, marker + color, markersize=markersize, **kwargs)

  # xlable
  if xlabel:
    plt.xlabel(xlabel)

  # ylabel
  if ylabel:
    plt.ylabel(ylabel)

  if xlim:
    plt.xlim(xlim[0], xlim[1])

  if ylim:
    plt.ylim(ylim[0], ylim[1])

  if title:
    plt.title(title)

  if show:
    plt.show()


def animate_2D(values,
               net_size,
               dt=None,
               val_min=None,
               val_max=None,
               cmap=None,
               frame_delay=10,
               frame_step=1,
               title_size=10,
               figsize=None,
               gif_dpi=None,
               video_fps=None,
               save_path=None,
               show=True):
  """Animate the potentials of the neuron group.

  Parameters
  ----------
  values : np.ndarray
      The membrane potentials of the neuron group.
  net_size : tuple
      The size of the neuron group.
  dt : float
      The time duration of each step.
  val_min : float, int
      The minimum of the potential.
  val_max : float, int
      The maximum of the potential.
  cmap : str
      The colormap.
  frame_delay : int, float
      The delay to show each frame.
  frame_step : int
      The step to show the potential. If `frame_step=3`, then each
      frame shows one of the every three steps.
  title_size : int
      The size of the title.
  figsize : None, tuple
      The size of the figure.
  gif_dpi : int
      Controls the dots per inch for the movie frames. This combined with
      the figure's size in inches controls the size of the movie. If
      ``None``, use defaults in matplotlib.
  video_fps : int
      Frames per second in the movie. Defaults to ``None``, which will use
      the animation's specified interval to set the frames per second.
  save_path : None, str
      The save path of the animation.
  show : bool
      Whether show the animation.

  Returns
  -------
  anim : animation.FuncAnimation
      The created animation function.
  """
  dt = math.get_dt() if dt is None else dt
  num_step, num_neuron = values.shape
  height, width = net_size

  values = np.asarray(values)
  val_min = values.min() if val_min is None else val_min
  val_max = values.max() if val_max is None else val_max

  figsize = figsize or (6, 6)

  fig = plt.figure(figsize=(figsize[0], figsize[1]), constrained_layout=True)
  gs = GridSpec(1, 1, figure=fig)
  fig.add_subplot(gs[0, 0])

  def frame(t):
    img = values[t]
    fig.clf()
    plt.pcolor(img, cmap=cmap, vmin=val_min, vmax=val_max)
    plt.colorbar()
    plt.axis('off')
    fig.suptitle(t="Time: {:.2f} ms".format((t + 1) * dt),
                 fontsize=title_size,
                 fontweight='bold')
    return [fig.gca()]

  values = values.reshape((num_step, height, width))
  anim = animation.FuncAnimation(fig=fig,
                                 func=frame,
                                 frames=list(range(1, num_step, frame_step)),
                                 init_func=None,
                                 interval=frame_delay,
                                 repeat_delay=3000)
  if save_path is None:
    if show:
      plt.show()
  else:
    logger.warning(f'Saving the animation into {save_path} ...')
    if save_path[-3:] == 'gif':
      anim.save(save_path, dpi=gif_dpi, writer='imagemagick')
    elif save_path[-3:] == 'mp4':
      anim.save(save_path, writer='ffmpeg', fps=video_fps, bitrate=3000)
    else:
      anim.save(save_path + '.mp4', writer='ffmpeg', fps=video_fps, bitrate=3000)
  return anim


def animate_1D(dynamical_vars,
               static_vars=(),
               dt=None,
               xlim=None,
               ylim=None,
               xlabel=None,
               ylabel=None,
               frame_delay=50.,
               frame_step=1,
               title_size=10,
               figsize=None,
               gif_dpi=None,
               video_fps=None,
               save_path=None,
               show=True,
               **kwargs):
  """Animation of one-dimensional data.

  Parameters
  ----------
  dynamical_vars : dict, np.ndarray, list of np.ndarray, list of dict
      The dynamical variables which will be animated.
  static_vars : dict, np.ndarray, list of np.ndarray, list of dict
      The static variables.
  xticks : list, np.ndarray
      The xticks.
  dt : float
      The numerical integration step.
  xlim : tuple
      The xlim.
  ylim : tuple
      The ylim.
  xlabel : str
      The xlabel.
  ylabel : str
      The ylabel.
  frame_delay : int, float
      The delay to show each frame.
  frame_step : int
      The step to show the potential. If `frame_step=3`, then each
      frame shows one of the every three steps.
  title_size : int
      The size of the title.
  figsize : None, tuple
      The size of the figure.
  gif_dpi : int
      Controls the dots per inch for the movie frames. This combined with
      the figure's size in inches controls the size of the movie. If
      ``None``, use defaults in matplotlib.
  video_fps : int
      Frames per second in the movie. Defaults to ``None``, which will use
      the animation's specified interval to set the frames per second.
  save_path : None, str
      The save path of the animation.
  show : bool
      Whether show the animation.

  Returns
  -------
  figure : plt.figure
      The created figure instance.
  """

  # check dt
  dt = math.get_dt() if dt is None else dt

  # check figure
  fig = plt.figure(figsize=(figsize or (6, 6)), constrained_layout=True)
  gs = GridSpec(1, 1, figure=fig)
  fig.add_subplot(gs[0, 0])

  # check dynamical variables
  final_dynamic_vars = []
  lengths = []
  has_legend = False
  if isinstance(dynamical_vars, (tuple, list)):
    for var in dynamical_vars:
      if isinstance(var, dict):
        assert 'ys' in var, 'Must provide "ys" item.'
        if 'legend' not in var:
          var['legend'] = None
        else:
          has_legend = True
        var['ys'] = np.asarray(var['ys'])
        if 'xs' not in var:
          var['xs'] = np.arange(var['ys'].shape[1])
      elif isinstance(var, (np.ndarray, math.ndarray)):
        var = np.asarray(var)
        var = {'ys': var,
               'xs': np.arange(var.shape[1]),
               'legend': None}
      else:
        raise ValueError(f'Unknown data type: {type(var)}')
      assert np.ndim(var['ys']) == 2, "Dynamic variable must be 2D data."
      lengths.append(var['ys'].shape[0])
      final_dynamic_vars.append(var)
  elif isinstance(dynamical_vars, dict):
    assert 'ys' in dynamical_vars, 'Must provide "ys" item.'
    if 'legend' not in dynamical_vars:
      dynamical_vars['legend'] = None
    else:
      has_legend = True
    dynamical_vars['ys'] = np.asarray(dynamical_vars['ys'])
    if 'xs' not in dynamical_vars:
      dynamical_vars['xs'] = np.arange(dynamical_vars['ys'].shape[1])
    lengths.append(dynamical_vars['ys'].shape[0])
    final_dynamic_vars.append(dynamical_vars)
  else:
    assert np.ndim(dynamical_vars) == 2, "Dynamic variable must be 2D data."
    dynamical_vars = np.asarray(dynamical_vars)
    lengths.append(dynamical_vars.shape[0])
    final_dynamic_vars.append({'ys': dynamical_vars,
                               'xs': np.arange(dynamical_vars.shape[1]),
                               'legend': None})
  lengths = np.array(lengths)
  assert np.all(lengths == lengths[0]), 'Dynamic variables must have equal length.'

  # check static variables
  final_static_vars = []
  if isinstance(static_vars, (tuple, list)):
    for var in static_vars:
      if isinstance(var, dict):
        assert 'data' in var, 'Must provide "ys" item.'
        if 'legend' not in var:
          var['legend'] = None
        else:
          has_legend = True
      elif isinstance(var, np.ndarray):
        var = {'data': var, 'legend': None}
      else:
        raise ValueError(f'Unknown data type: {type(var)}')
      assert np.ndim(var['data']) == 1, "Static variable must be 1D data."
      final_static_vars.append(var)
  elif isinstance(static_vars, np.ndarray):
    final_static_vars.append({'data': static_vars,
                              'xs': np.arange(static_vars.shape[0]),
                              'legend': None})
  elif isinstance(static_vars, dict):
    assert 'ys' in static_vars, 'Must provide "ys" item.'
    if 'legend' not in static_vars:
      static_vars['legend'] = None
    else:
      has_legend = True
    if 'xs' not in static_vars:
      static_vars['xs'] = np.arange(static_vars['ys'].shape[0])
    final_static_vars.append(static_vars)

  else:
    raise ValueError(f'Unknown static data type: {type(static_vars)}')

  # ylim
  if ylim is None:
    ylim_min = np.inf
    ylim_max = -np.inf
    for var in final_dynamic_vars + final_static_vars:
      if var['ys'].max() > ylim_max:
        ylim_max = var['ys'].max()
      if var['ys'].min() < ylim_min:
        ylim_min = var['ys'].min()
    if ylim_min > 0:
      ylim_min = ylim_min * 0.98
    else:
      ylim_min = ylim_min * 1.02
    if ylim_max > 0:
      ylim_max = ylim_max * 1.02
    else:
      ylim_max = ylim_max * 0.98
    ylim = (ylim_min, ylim_max)

  def frame(t):
    fig.clf()
    for dvar in final_dynamic_vars:
      plt.plot(dvar['xs'], dvar['ys'][t], label=dvar['legend'], **kwargs)
    for svar in final_static_vars:
      plt.plot(svar['xs'], svar['ys'], label=svar['legend'], **kwargs)
    if xlim is not None:
      plt.xlim(xlim[0], xlim[1])
    if has_legend:
      plt.legend()
    if xlabel:
      plt.xlabel(xlabel)
    if ylabel:
      plt.ylabel(ylabel)
    plt.ylim(ylim[0], ylim[1])
    fig.suptitle(t="Time: {:.2f} ms".format((t + 1) * dt),
                 fontsize=title_size,
                 fontweight='bold')
    return [fig.gca()]

  anim_result = animation.FuncAnimation(fig=fig,
                                        func=frame,
                                        frames=range(1, lengths[0], frame_step),
                                        init_func=None,
                                        interval=frame_delay,
                                        repeat_delay=3000)

  # save or show
  if save_path is None:
    if show: plt.show()
  else:
    logger.warning(f'Saving the animation into {save_path} ...')
    if save_path[-3:] == 'gif':
      anim_result.save(save_path, dpi=gif_dpi, writer='imagemagick')
    elif save_path[-3:] == 'mp4':
      anim_result.save(save_path, writer='ffmpeg', fps=video_fps, bitrate=3000)
    else:
      anim_result.save(save_path + '.mp4', writer='ffmpeg', fps=video_fps, bitrate=3000)
  return fig



# -*- coding: utf-8 -*-

try:
  from numba import njit
except (ImportError, ModuleNotFoundError):
  njit = None


__all__ = [
  'numba_jit'
]


def numba_jit(f=None, **kwargs):
  if f is None:
    return lambda f: (f if (njit is None) else njit(f, **kwargs))
  else:
    if njit is None:
      return f
    else:
      return njit(f)




# -*- coding: utf-8 -*-

import warnings

import jax.numpy as jnp
import numpy as onp

import brainpy.math as bm
from brainpy.base.base import Base

__all__ = [
  'Module',
]


def _check_args(args):
  if args is None:
    return tuple()
  elif isinstance(args, tuple):
    return args
  else:
    return (args,)


class Module(Base):
  """Basic module class.

  .. deprecated:: 2.1.0
  """

  @staticmethod
  def get_param(param, size):
    return bm.TrainVar(Module.init_param(param, size))

  @staticmethod
  def init_param(param, size):
    if param is None:
      return None
    if callable(param):
      param = param(size)
    elif isinstance(param, onp.ndarray):
      param = bm.asarray(param)
    elif isinstance(param, (bm.JaxArray, jnp.ndarray)):
      pass
    else:
      raise ValueError(f'Unknown param type {type(param)}: {param}')
    assert param.shape == size, f'"param.shape" is not the required size {size}'
    return param

  def __init__(self, name=None):  # initialize parameters
    warnings.warn('Please use "brainpy.rnns.Module" instead. '
                  '"brainpy.layers.Module" is deprecated since '
                  'version 2.1.0.', DeprecationWarning)
    super(Module, self).__init__(name=name)

  def __call__(self, *args, **kwargs):  # initialize variables
    return self.call(*args, **kwargs)

  def call(self, *args, **kwargs):
    raise NotImplementedError




# -*- coding: utf-8 -*-

from jax.numpy import linalg

from brainpy.math.jaxarray import JaxArray
from brainpy.math.numpy_ops import _remove_jaxarray

__all__ = [
  'cholesky', 'cond', 'det', 'eig', 'eigh', 'eigvals', 'eigvalsh', 'inv', 'svd',
  'lstsq', 'matrix_power', 'matrix_rank', 'norm', 'pinv', 'qr', 'solve', 'slogdet',
  'tensorinv', 'tensorsolve', 'multi_dot'
]


def cholesky(a):
  a = _remove_jaxarray(a)
  return JaxArray(linalg.cholesky(a))


def cond(x, p=None):
  x = _remove_jaxarray(x)
  p = _remove_jaxarray(p)
  return linalg.cond(x, p=p)


def det(a):
  a = _remove_jaxarray(a)
  return JaxArray(linalg.det(a))


def eig(a):
  a = _remove_jaxarray(a)
  w, v = linalg.eig(a)
  return JaxArray(w), JaxArray(v)


def eigh(a, UPLO='L'):
  a = _remove_jaxarray(a)
  w, v = linalg.eigh(a, UPLO=UPLO)
  return JaxArray(w), JaxArray(v)


def eigvals(a):
  a = _remove_jaxarray(a)
  return JaxArray(linalg.eigvals(a))


def eigvalsh(a, UPLO='L'):
  a = _remove_jaxarray(a)
  return JaxArray(linalg.eigvalsh(a, UPLO=UPLO))


def inv(a):
  a = _remove_jaxarray(a)
  return JaxArray(linalg.inv(a))


def svd(a, full_matrices=True, compute_uv=True):
  a = _remove_jaxarray(a)
  u, s, vh = linalg.svd(a, full_matrices=full_matrices, compute_uv=compute_uv)
  return JaxArray(u), JaxArray(s), JaxArray(vh)


def lstsq(a, b, rcond=None):
  a = _remove_jaxarray(a)
  b = _remove_jaxarray(b)
  x, resid, rank, s = linalg.lstsq(a, b, rcond=rcond)
  return JaxArray(x), JaxArray(resid), rank, JaxArray(s)


def matrix_power(a, n):
  a = _remove_jaxarray(a)
  return JaxArray(linalg.matrix_power(a, n))


def matrix_rank(M, tol=None):
  M = _remove_jaxarray(M)
  r = linalg.matrix_rank(M, tol=tol)
  return r if isinstance(r, int) else JaxArray(r)


def norm(x, ord=None, axis=None, keepdims=False):
  x = _remove_jaxarray(x)
  r = linalg.norm(x, ord=ord, axis=axis, keepdims=keepdims)
  return r if axis is None else JaxArray(r)


def pinv(a, rcond=None):
  a = _remove_jaxarray(a)
  if isinstance(rcond, JaxArray): rcond = rcond.value
  return JaxArray(linalg.pinv(a, rcond=rcond))


def qr(a, mode="reduced"):
  a = _remove_jaxarray(a)
  r = linalg.qr(a, mode=mode)
  if isinstance(r, tuple):
    return JaxArray(r[0]), JaxArray(r[1])
  else:
    return JaxArray(r)


def solve(a, b):
  a = _remove_jaxarray(a)
  b = _remove_jaxarray(b)
  return JaxArray(linalg.solve(a, b))


def slogdet(a):
  a = _remove_jaxarray(a)
  return JaxArray(linalg.slogdet(a))


def tensorinv(a, ind=2):
  a = _remove_jaxarray(a)
  return JaxArray(linalg.tensorinv(a, ind=ind))


def tensorsolve(a, b, axes=None):
  a = _remove_jaxarray(a)
  b = _remove_jaxarray(b)
  return JaxArray(linalg.tensorsolve(a, b, axes=axes))


def multi_dot(arrays):
  arrays = [_remove_jaxarray(a) for a in arrays]
  return JaxArray(linalg.multi_dot(arrays))



# -*- coding: utf-8 -*-

from typing import Dict, Optional, Any

from brainpy.math import activations
from brainpy.nn.base import Node

__all__ = [
  'Activation'
]


class Activation(Node):
  """Activation node.

  Parameters
  ----------
  activation : str
    The name of the activation function.
  fun_setting : optional, dict
    The settings for the activation function.
  """

  def __init__(self,
               activation: str = 'relu',
               fun_setting: Optional[Dict[str, Any]] = None,
               trainable: bool = False,
               name: str = None,
               **kwargs):
    if name is None:
      name = self.unique_name(type_=f'{activation}_activation')
    super(Activation, self).__init__(name=name, trainable=trainable, **kwargs)

    self._activation = activations.get(activation)
    self._fun_setting = dict() if (fun_setting is None) else fun_setting
    assert isinstance(self._fun_setting, dict), '"fun_setting" must be a dict.'

  def init_ff_conn(self):
    self.set_output_shape(self.feedforward_shapes)

  def forward(self, ff, **shared_kwargs):
    return self._activation(ff, **self._fun_setting)



# -*- coding: utf-8 -*-

from functools import partial

import jax.numpy as jnp
from jax import vmap
import numpy as np

import brainpy.math as bm
from brainpy import errors
from brainpy.analysis import stability, utils, constants as C
from brainpy.analysis.lowdim.lowdim_analyzer import *

pyplot = None

__all__ = [
  'Bifurcation1D',
  'Bifurcation2D',
  'FastSlow1D',
  'FastSlow2D',
]


class Bifurcation1D(Num1DAnalyzer):
  """Bifurcation analysis of 1D system.

  Using this class, we can make co-dimension1 or co-dimension2 bifurcation analysis.
  """

  def __init__(self, model, target_pars, target_vars, fixed_vars=None,
               pars_update=None, resolutions=None, options=None):
    super(Bifurcation1D, self).__init__(model=model,
                                        target_pars=target_pars,
                                        target_vars=target_vars,
                                        fixed_vars=fixed_vars,
                                        pars_update=pars_update,
                                        resolutions=resolutions,
                                        options=options)

    if len(self.target_pars) == 0:
      raise ValueError

  @property
  def F_vmap_dfxdx(self):
    if C.F_vmap_dfxdx not in self.analyzed_results:
      f = bm.jit(vmap(bm.vector_grad(self.F_fx, argnums=0)), device=self.jit_device)
      self.analyzed_results[C.F_vmap_dfxdx] = f
    return self.analyzed_results[C.F_vmap_dfxdx]

  def plot_bifurcation(self, with_plot=True, show=False, with_return=False,
                       tol_aux=1e-8, loss_screen=None):
    global pyplot
    if pyplot is None: from matplotlib import pyplot
    utils.output('I am making bifurcation analysis ...')

    xs = self.resolutions[self.x_var]
    vps = bm.meshgrid(xs, *tuple(self.resolutions[p] for p in self.target_par_names))
    vps = tuple(jnp.moveaxis(vp.value, 0, 1).flatten() for vp in vps)
    candidates = vps[0]
    pars = vps[1:]
    fixed_points, _, pars = self._get_fixed_points(candidates, *pars,
                                                   tol_aux=tol_aux,
                                                   loss_screen=loss_screen,
                                                   num_seg=len(xs))
    dfxdx = np.asarray(self.F_vmap_dfxdx(jnp.asarray(fixed_points), *pars))
    pars = tuple(np.asarray(p) for p in pars)

    if with_plot:
      if len(self.target_pars) == 1:
        container = {c: {'p': [], 'x': []} for c in stability.get_1d_stability_types()}

        # fixed point
        for p, x, dx in zip(pars[0], fixed_points, dfxdx):
          fp_type = stability.stability_analysis(dx)
          container[fp_type]['p'].append(p)
          container[fp_type]['x'].append(x)

        # visualization
        pyplot.figure(self.x_var)
        for fp_type, points in container.items():
          if len(points['x']):
            plot_style = stability.plot_scheme[fp_type]
            pyplot.plot(points['p'], points['x'], '.', **plot_style, label=fp_type)
        pyplot.xlabel(self.target_par_names[0])
        pyplot.ylabel(self.x_var)

        scale = (self.lim_scale - 1) / 2
        pyplot.xlim(*utils.rescale(self.target_pars[self.target_par_names[0]], scale=scale))
        pyplot.ylim(*utils.rescale(self.target_vars[self.x_var], scale=scale))

        pyplot.legend()
        if show:
          pyplot.show()

      elif len(self.target_pars) == 2:
        container = {c: {'p0': [], 'p1': [], 'x': []} for c in stability.get_1d_stability_types()}

        # fixed point
        for p0, p1, x, dx in zip(pars[0], pars[1], fixed_points, dfxdx):
          fp_type = stability.stability_analysis(dx)
          container[fp_type]['p0'].append(p0)
          container[fp_type]['p1'].append(p1)
          container[fp_type]['x'].append(x)

        # visualization
        fig = pyplot.figure(self.x_var)
        ax = fig.add_subplot(projection='3d')
        for fp_type, points in container.items():
          if len(points['x']):
            plot_style = stability.plot_scheme[fp_type]
            xs = points['p0']
            ys = points['p1']
            zs = points['x']
            ax.scatter(xs, ys, zs, **plot_style, label=fp_type)

        ax.set_xlabel(self.target_par_names[0])
        ax.set_ylabel(self.target_par_names[1])
        ax.set_zlabel(self.x_var)

        scale = (self.lim_scale - 1) / 2
        ax.set_xlim(*utils.rescale(self.target_pars[self.target_par_names[0]], scale=scale))
        ax.set_ylim(*utils.rescale(self.target_pars[self.target_par_names[1]], scale=scale))
        ax.set_zlim(*utils.rescale(self.target_vars[self.x_var], scale=scale))

        ax.grid(True)
        ax.legend()
        if show:
          pyplot.show()

      else:
        raise errors.BrainPyError(f'Cannot visualize co-dimension {len(self.target_pars)} '
                                  f'bifurcation.')
    if with_return:
      return fixed_points, pars, dfxdx


class Bifurcation2D(Num2DAnalyzer):
  """Bifurcation analysis of 2D system.

  Using this class, we can make co-dimension1 or co-dimension2 bifurcation analysis.
  """

  def __init__(self, model, target_pars, target_vars, fixed_vars=None,
               pars_update=None, resolutions=None, options=None):
    super(Bifurcation2D, self).__init__(model=model,
                                        target_pars=target_pars,
                                        target_vars=target_vars,
                                        fixed_vars=fixed_vars,
                                        pars_update=pars_update,
                                        resolutions=resolutions,
                                        options=options)

    if len(self.target_pars) == 0:
      raise ValueError

    self._fixed_points = None

  @property
  def F_vmap_jacobian(self):
    if C.F_vmap_jacobian not in self.analyzed_results:
      f1 = lambda xy, *args: jnp.array([self.F_fx(xy[0], xy[1], *args),
                                        self.F_fy(xy[0], xy[1], *args)])
      f2 = bm.jit(vmap(bm.jacobian(f1)), device=self.jit_device)
      self.analyzed_results[C.F_vmap_jacobian] = f2
    return self.analyzed_results[C.F_vmap_jacobian]

  def plot_bifurcation(self, with_plot=True, show=False, with_return=False,
                       tol_aux=1e-8, tol_unique=1e-2, tol_opt_candidate=None,
                       num_par_segments=1, num_fp_segment=1, nullcline_aux_filter=1.,
                       select_candidates='aux_rank', num_rank=100):
    """Make the bifurcation analysis.

    Parameters
    ----------
    with_plot: bool
      Whether plot the bifurcation figure.
    show: bool
      Whether show the figure.
    with_return: bool
      Whether return the computed bifurcation results.
    tol_aux: float
      The loss tolerance of auxiliary function :math:`f_{aux}` to confirm the fixed
      point. Default is 1e-7. Once :math:`f_{aux}(x_1) < \mathrm{tol\_aux}`,
      :math:`x_1` will be a fixed point.
    tol_unique: float
      The tolerance of distance between candidate fixed points to confirm they are
      the same. Default is 1e-2. If :math:`|x_1 - x_2| > \mathrm{tol\_unique}`,
      then :math:`x_1` and :math:`x_2` are unique fixed points. Otherwise,
      :math:`x_1` and :math:`x_2` will be treated as a same fixed point.
    tol_opt_candidate: float, optional
      The tolerance of auxiliary function :math:`f_{aux}` to select candidate
      initial points for fixed point optimization.
    num_par_segments: int, sequence of int
      How to segment parameters.
    num_fp_segment: int
      How to segment fixed points.
    nullcline_aux_filter: float
      The
    select_candidates: str
      The method to select candidate fixed points. It can be:

      - ``fx-nullcline``: use the points of fx-nullcline.
      - ``fy-nullcline``: use the points of fy-nullcline.
      - ``nullclines``: use the points in both of fx-nullcline and fy-nullcline.
      - ``aux_rank``: use the minimal value of points for the auxiliary function.
    num_rank: int
      The number of candidates to be used to optimize the fixed points.
      rank to use.

    Returns
    -------
    results : tuple
      Return a tuple of analyzed results:

      - fixed points: a 2D matrix with the shape of (num_point, num_var)
      - parameters: a 2D matrix with the shape of (num_point, num_par)
      - jacobians: a 3D tensors with the shape of (num_point, 2, 2)
    """
    global pyplot
    if pyplot is None: from matplotlib import pyplot
    utils.output('I am making bifurcation analysis ...')

    if self._can_convert_to_one_eq():
      if self.convert_type() == C.x_by_y:
        X = self.resolutions[self.y_var].value
      else:
        X = self.resolutions[self.x_var].value
      pars = tuple(self.resolutions[p].value for p in self.target_par_names)
      mesh_values = jnp.meshgrid(*((X,) + pars))
      mesh_values = tuple(jnp.moveaxis(v, 0, 1).flatten() for v in mesh_values)
      candidates = mesh_values[0]
      parameters = mesh_values[1:]

    else:
      if select_candidates == 'fx-nullcline':
        fx_nullclines = self._get_fx_nullcline_points(num_segments=num_par_segments,
                                                      fp_aux_filter=nullcline_aux_filter)
        candidates = fx_nullclines[0]
        parameters = fx_nullclines[1:]
      elif select_candidates == 'fy-nullcline':
        fy_nullclines = self._get_fy_nullcline_points(num_segments=num_par_segments,
                                                      fp_aux_filter=nullcline_aux_filter)
        candidates = fy_nullclines[0]
        parameters = fy_nullclines[1:]
      elif select_candidates == 'nullclines':
        fx_nullclines = self._get_fx_nullcline_points(num_segments=num_par_segments,
                                                      fp_aux_filter=nullcline_aux_filter)
        fy_nullclines = self._get_fy_nullcline_points(num_segments=num_par_segments,
                                                      fp_aux_filter=nullcline_aux_filter)
        candidates = jnp.vstack([fx_nullclines[0], fy_nullclines[0]])
        parameters = [jnp.concatenate([fx_nullclines[i], fy_nullclines[i]])
                      for i in range(1, len(fy_nullclines))]
      elif select_candidates == 'aux_rank':
        assert nullcline_aux_filter > 0.
        candidates, parameters = self._get_fp_candidates_by_aux_rank(num_segments=num_par_segments,
                                                                     num_rank=num_rank)
      else:
        raise ValueError
    candidates, _, parameters = self._get_fixed_points(candidates,
                                                       *parameters,
                                                       tol_aux=tol_aux,
                                                       tol_unique=tol_unique,
                                                       tol_opt_candidate=tol_opt_candidate,
                                                       num_segment=num_fp_segment)
    candidates = np.asarray(candidates)
    parameters = np.stack(tuple(np.asarray(p) for p in parameters)).T
    utils.output('I am trying to filter out duplicate fixed points ...')
    final_fps = []
    final_pars = []
    for par in np.unique(parameters, axis=0):
      ids = np.where(np.all(parameters == par, axis=1))[0]
      fps, ids2 = utils.keep_unique(candidates[ids], tolerance=tol_unique)
      final_fps.append(fps)
      final_pars.append(parameters[ids[ids2]])
    final_fps = np.vstack(final_fps)  # with the shape of (num_point, num_var)
    final_pars = np.vstack(final_pars)  # with the shape of (num_point, num_par)
    jacobians = np.asarray(self.F_vmap_jacobian(jnp.asarray(final_fps), *final_pars.T))
    utils.output(f'{C.prefix}Found {len(final_fps)} fixed points.')

    # remember the fixed points for later limit cycle plotting
    self._fixed_points = (final_fps, final_pars)

    if with_plot:
      # bifurcation analysis of co-dimension 1
      if len(self.target_pars) == 1:
        container = {c: {'p': [], self.x_var: [], self.y_var: []}
                     for c in stability.get_2d_stability_types()}

        # fixed point
        for p, xy, J in zip(final_pars, final_fps, jacobians):
          fp_type = stability.stability_analysis(J)
          container[fp_type]['p'].append(p[0])
          container[fp_type][self.x_var].append(xy[0])
          container[fp_type][self.y_var].append(xy[1])

        # visualization
        for var in self.target_var_names:
          pyplot.figure(var)
          for fp_type, points in container.items():
            if len(points['p']):
              plot_style = stability.plot_scheme[fp_type]
              pyplot.plot(points['p'], points[var], '.', **plot_style, label=fp_type)
          pyplot.xlabel(self.target_par_names[0])
          pyplot.ylabel(var)

          scale = (self.lim_scale - 1) / 2
          pyplot.xlim(*utils.rescale(self.target_pars[self.target_par_names[0]], scale=scale))
          pyplot.ylim(*utils.rescale(self.target_vars[var], scale=scale))

          pyplot.legend()
        if show:
          pyplot.show()

      # bifurcation analysis of co-dimension 2
      elif len(self.target_pars) == 2:
        container = {c: {'p0': [], 'p1': [], self.x_var: [], self.y_var: []}
                     for c in stability.get_2d_stability_types()}

        # fixed point
        for p, xy, J in zip(final_pars, final_fps, jacobians):
          fp_type = stability.stability_analysis(J)
          container[fp_type]['p0'].append(p[0])
          container[fp_type]['p1'].append(p[1])
          container[fp_type][self.x_var].append(xy[0])
          container[fp_type][self.y_var].append(xy[1])

        # visualization
        for var in self.target_var_names:
          fig = pyplot.figure(var)
          ax = fig.add_subplot(projection='3d')
          for fp_type, points in container.items():
            if len(points['p0']):
              plot_style = stability.plot_scheme[fp_type]
              xs = points['p0']
              ys = points['p1']
              zs = points[var]
              ax.scatter(xs, ys, zs, **plot_style, label=fp_type)

          ax.set_xlabel(self.target_par_names[0])
          ax.set_ylabel(self.target_par_names[1])
          ax.set_zlabel(var)
          scale = (self.lim_scale - 1) / 2
          ax.set_xlim(*utils.rescale(self.target_pars[self.target_par_names[0]], scale=scale))
          ax.set_ylim(*utils.rescale(self.target_pars[self.target_par_names[1]], scale=scale))
          ax.set_zlim(*utils.rescale(self.target_vars[var], scale=scale))
          ax.grid(True)
          ax.legend()
        if show:
          pyplot.show()

      else:
        raise ValueError('Unknown length of parameters.')

    if with_return:
      return final_fps, final_pars, jacobians

  def plot_limit_cycle_by_sim(self, duration=100, with_plot=True, with_return=False,
                              plot_style=None, tol=0.001, show=False, dt=None, offset=1.):
    global pyplot
    if pyplot is None: from matplotlib import pyplot
    utils.output('I am plotting the limit cycle ...')
    if self._fixed_points is None:
      utils.output('No fixed points found, you may call "plot_bifurcation(with_plot=True)" first.')
      return

    final_fps, final_pars = self._fixed_points
    dt = bm.get_dt() if dt is None else dt
    traject_model = utils.TrajectModel(
      initial_vars={self.x_var: final_fps[:, 0] + offset, self.y_var: final_fps[:, 1] + offset},
      integrals={self.x_var: self.F_int_x, self.y_var: self.F_int_y},
      pars={p: v for p, v in zip(self.target_par_names, final_pars.T)},
      dt=dt
    )
    mon_res = traject_model.run(duration=duration)

    # find limit cycles
    vs_limit_cycle = tuple({'min': [], 'max': []} for _ in self.target_var_names)
    ps_limit_cycle = tuple([] for _ in self.target_par_names)
    for i in range(mon_res[self.x_var].shape[1]):
      data = mon_res[self.x_var][:, i]
      max_index = utils.find_indexes_of_limit_cycle_max(data, tol=tol)
      if max_index[0] != -1:
        cycle = data[max_index[0]: max_index[1]]
        vs_limit_cycle[0]['max'].append(mon_res[self.x_var][max_index[1], i])
        vs_limit_cycle[0]['min'].append(cycle.min())
        cycle = mon_res[self.y_var][max_index[0]: max_index[1], i]
        vs_limit_cycle[1]['max'].append(mon_res[self.y_var][max_index[1], i])
        vs_limit_cycle[1]['min'].append(cycle.min())
        for j in range(len(self.target_par_names)):
          ps_limit_cycle[j].append(final_pars[i, j])
    vs_limit_cycle = tuple({k: np.asarray(v) for k, v in lm.items()} for lm in vs_limit_cycle)
    ps_limit_cycle = tuple(np.array(p) for p in ps_limit_cycle)

    # visualization
    if with_plot:
      if plot_style is None: plot_style = dict()
      fmt = plot_style.pop('fmt', '*')

      if len(self.target_par_names) == 2:
        if len(ps_limit_cycle[0]):
          for i, var in enumerate(self.target_var_names):
            pyplot.figure(var)
            pyplot.plot(ps_limit_cycle[0], ps_limit_cycle[1], vs_limit_cycle[i]['max'],
                        **plot_style, label='limit cycle (max)')
            pyplot.plot(ps_limit_cycle[0], ps_limit_cycle[1], vs_limit_cycle[i]['min'],
                        **plot_style, label='limit cycle (min)')
            pyplot.legend()

      elif len(self.target_par_names) == 1:
        if len(ps_limit_cycle[0]):
          for i, var in enumerate(self.target_var_names):
            pyplot.figure(var)
            pyplot.plot(ps_limit_cycle[0], vs_limit_cycle[i]['max'], fmt,
                        **plot_style, label='limit cycle (max)')
            pyplot.plot(ps_limit_cycle[0], vs_limit_cycle[i]['min'], fmt,
                        **plot_style, label='limit cycle (min)')
            pyplot.legend()

      else:
        raise errors.AnalyzerError

      if show:
        pyplot.show()

    if with_return:
      return vs_limit_cycle, ps_limit_cycle


class FastSlow1D(Bifurcation1D):
  def __init__(self, model, fast_vars, slow_vars, fixed_vars=None,
               pars_update=None, resolutions=None, options=None):
    super(FastSlow1D, self).__init__(model=model,
                                     target_pars=slow_vars,
                                     target_vars=fast_vars,
                                     fixed_vars=fixed_vars,
                                     pars_update=pars_update,
                                     resolutions=resolutions,
                                     options=options)

    # standard integrators
    self._std_integrators = dict()
    for key, intg in self.model.name2integral.items():
      wrap_x = utils.std_derivative(utils.get_args(self.model.name2derivative[key])[1],
                                    self.target_var_names + self.target_par_names, [])
      self._std_integrators[key] = partial(wrap_x(self.model.name2integral[key]),
                                           **(self.pars_update + self.fixed_vars))

  def plot_trajectory(self, initials, duration, plot_durations=None,
                      dt=None, show=False, with_plot=True, with_return=False):
    global pyplot
    if pyplot is None: from matplotlib import pyplot
    utils.output('I am plotting the trajectory ...')

    # check the initial values
    initials = utils.check_initials(initials, self.target_var_names + self.target_par_names)

    # 2. format the running duration
    assert isinstance(duration, (int, float))

    # 3. format the plot duration
    plot_durations = utils.check_plot_durations(plot_durations, duration, initials)

    # 5. run the network
    dt = bm.get_dt() if dt is None else dt

    traject_model = utils.TrajectModel(initial_vars=initials, integrals=self._std_integrators, dt=dt)
    mon_res = traject_model.run(duration=duration)

    if with_plot:
      assert len(self.target_par_names) <= 2

      # plots
      for i, initial in enumerate(zip(*list(initials.values()))):
        # legend
        legend = f'$traj_{i}$: '
        for j, key in enumerate(self.target_var_names):
          legend += f'{key}={initial[j]}, '
        legend = legend[:-2]

        # visualization
        start = int(plot_durations[i][0] / dt)
        end = int(plot_durations[i][1] / dt)
        p1_var = self.target_par_names[0]
        if len(self.target_par_names) == 1:
          lines = pyplot.plot(mon_res[self.x_var][start: end, i],
                              mon_res[p1_var][start: end, i], label=legend)
        elif len(self.target_par_names) == 2:
          p2_var = self.target_par_names[1]
          lines = pyplot.plot(mon_res[self.x_var][start: end, i],
                              mon_res[p1_var][start: end, i],
                              mon_res[p2_var][start: end, i],
                              label=legend)
        else:
          raise ValueError
        utils.add_arrow(lines[0])

      # # visualization of others
      # plt.xlabel(self.x_var)
      # plt.ylabel(self.target_par_names[0])
      # scale = (self.lim_scale - 1.) / 2
      # plt.xlim(*utils.rescale(self.target_vars[self.x_var], scale=scale))
      # plt.ylim(*utils.rescale(self.target_vars[self.target_par_names[0]], scale=scale))
      pyplot.legend()

      if show:
        pyplot.show()

    if with_return:
      return mon_res


class FastSlow2D(Bifurcation2D):
  def __init__(self, model, fast_vars, slow_vars, fixed_vars=None,
               pars_update=None, resolutions=0.1, options=None):
    super(FastSlow2D, self).__init__(model=model,
                                     target_pars=slow_vars,
                                     target_vars=fast_vars,
                                     fixed_vars=fixed_vars,
                                     pars_update=pars_update,
                                     resolutions=resolutions,
                                     options=options)
    # standard integrators
    self._std_integrators = dict()
    for key, intg in self.model.name2integral.items():
      wrap_x = utils.std_derivative(utils.get_args(self.model.name2derivative[key])[1],
                                    self.target_var_names + self.target_par_names, [])
      self._std_integrators[key] = partial(wrap_x(self.model.name2integral[key]),
                                           **(self.pars_update + self.fixed_vars))

  def plot_trajectory(self, initials, duration, plot_durations=None,
                      dt=None, show=False, with_plot=True, with_return=False):
    global pyplot
    if pyplot is None: from matplotlib import pyplot
    utils.output('I am plotting the trajectory ...')

    # check the initial values
    initials = utils.check_initials(initials, self.target_var_names + self.target_par_names)

    # 2. format the running duration
    assert isinstance(duration, (int, float))

    # 3. format the plot duration
    plot_durations = utils.check_plot_durations(plot_durations, duration, initials)

    # 5. run the network
    dt = bm.get_dt() if dt is None else dt

    traject_model = utils.TrajectModel(initial_vars=initials, integrals=self._std_integrators, dt=dt)
    mon_res = traject_model.run(duration=duration)

    if with_plot:
      assert len(self.target_par_names) <= 1
      # plots
      for i, initial in enumerate(zip(*list(initials.values()))):
        # legend
        legend = f'$traj_{i}$: '
        for j, key in enumerate(self.target_var_names):
          legend += f'{key}={initial[j]}, '
        legend = legend[:-2]

        start = int(plot_durations[i][0] / dt)
        end = int(plot_durations[i][1] / dt)

        # visualization
        pyplot.figure(self.x_var)
        lines = pyplot.plot(mon_res[self.target_par_names[0]][start: end, i],
                            mon_res[self.x_var][start: end, i],
                            label=legend)
        utils.add_arrow(lines[0])

        pyplot.figure(self.y_var)
        lines = pyplot.plot(mon_res[self.target_par_names[0]][start: end, i],
                            mon_res[self.y_var][start: end, i],
                            label=legend)
        utils.add_arrow(lines[0])

      pyplot.figure(self.x_var)
      pyplot.legend()
      pyplot.figure(self.y_var)
      pyplot.legend()

      if show:
        pyplot.show()

    if with_return:
      return mon_res
